import argparse
import logging
from pathlib import Path
import pyaxml
from pyaxml import conf
try:
    from lxml import etree
except ImportError:
    import xml.etree.ElementTree as etree


def main() -> int:
    """cli function"""

    parser = argparse.ArgumentParser()
    parser.add_argument('command', choices=[
        "axml2xml",
        "xml2axml",
        "arsc2xml"
    ]) 
    parser.add_argument("-i", "--input", help="Specify the apk input file")
    parser.add_argument("-o", "--output", help="Specify the apk output file")
    parser.add_argument("-v", "--version", help="version of pyaxml", action="store_true")

    args = parser.parse_args()

    if args.version:
        print(f"version {conf.VERSION}")
        return 0
    
    if not args.input:
        logging.error("No input provided")
        parser.print_help()
        return 1
    
    if args.command == "axml2xml":
        with open(args.input, "rb") as f:
            # Read AXML
            axml, _ = pyaxml.AXML.from_axml(f.read())
            xml = axml.to_xml()
        # Rewrite the file
        if args.output:
            with open(args.output, "w") as f:
                f.write(etree.tostring(xml, encoding='unicode', pretty_print=True))
        else:
            print(etree.tostring(xml, encoding='unicode', pretty_print=True))
    elif args.command == "xml2axml":
        if not args.output:
            logging.error("No output provided")
            parser.print_help()
            return 1
        with open(args.input, "r") as f:
            # Read XML
            root = etree.fromstring(f.read())
            axml_object = pyaxml.AXML()
            axml_object.from_xml(root)
        # Rewrite the file
        with open(args.output, "wb") as f:
            f.write(axml_object.pack())
    elif args.command == "arsc2xml":
        with open(args.input, "rb") as f:
            # Read AXML
            axml, _ = pyaxml.ARSC.from_axml(f.read())
            xml = axml.list_packages()
        # Rewrite the file
        if args.output:
            with open(args.output, "w") as f:
                f.write(xml)
        else:
            print(xml)



import unittest
from struct import pack
from pyaxml.arscobject import ARSC
from pyaxml.stringblocks import StringBlocks
from pyaxml.proto import axml_pb2


class TestARSC(unittest.TestCase):
    def setUp(self):
        """Set up a basic ARSC instance with real axml_pb2 data."""
        # Create a proto instance for the test
        self.proto = axml_pb2.ARSC()

        # Populate proto with basic test data
        self.proto.header_res.hnd.size = 0
        self.proto.header_res.package_count = 1

        # Add a simple package
        package = self.proto.restablespackage.add()
        package.name = "test_package\x00"

        str_blocks = StringBlocks()
        str_blocks.get("string")
        package.type_sp_string.CopyFrom(str_blocks.proto)

        str_blocks = StringBlocks()
        str_blocks.get("name")
        package.key_sp_string.CopyFrom(str_blocks.proto)

        str_blocks = StringBlocks()
        str_blocks.get("test1")
        str_blocks.get("value")
        str_blocks.get("val")
        str_blocks.get("test4")

        self.proto.stringblocks.CopyFrom(str_blocks.proto)
        # Add type and key string blocks to the package
        # package.type_sp_string.append("string\x00")
        # package.key_sp_string.append("name\x00")

        # Add a simple resource type
        restype = package.restypes.add()
        restype.hnd.type = axml_pb2.ResType.RES_TABLE_TYPE_TYPE
        restype.typetype.id = 1
        entry = restype.typetype.tables.add()
        entry.present = True
        entry.index = 0
        entry.key.data = 1  # Simulated value
        entry.key.data_type = 3  # Simulate a string type

        self.arsc = ARSC(proto=self.proto)

    def test_get_packages(self):
        """Test get_packages method."""
        packages = self.arsc.get_packages()
        self.assertEqual(packages, ["test_package"])

    def test_get_value(self):
        """Test get_value method."""

        # Fetch the value using the ARSC method
        value = self.arsc.get_value("test_package", 0x7F010000)
        self.assertEqual(value, "value")

    def test_get_id(self):
        """Test get_id method."""
        type_, key, rid = self.arsc.get_id("test_package", 0x7F010000)
        self.assertEqual(type_, "string")
        self.assertEqual(key, "name")
        self.assertEqual(rid, 0x7F010000)

    def test_get_resource_xml_name(self):
        """Test get_resource_xml_name method."""
        xml_name = self.arsc.get_resource_xml_name(0x7F010000, "test_package")
        self.assertEqual(xml_name, "@string/name")

    def test_add_id_public(self):
        """Test adding a new ID to the public strings."""
        new_id = self.arsc.add_id_public(
            "test_package", "string", "new_name", "new_value"
        )
        self.assertIsNotNone(new_id)

    def test_list_packages(self):
        """Test list_packages method."""
        output = self.arsc.list_packages()
        self.assertIn('type="string"', output)
        self.assertIn('name="name"', output)


if __name__ == "__main__":
    unittest.main()



import os
import unittest

import io
from xml.etree import ElementTree

import pyaxml

test_dir = os.path.dirname(os.path.abspath(__file__))


def is_valid_manifest(tree):
    # We can not really check much more...
    if tree.tag == "manifest" and "package" in tree.attrib:
        return True
    return False


# text_compare and xml_compare are modified from
# https://bitbucket.org/ianb/formencode/src/tip/formencode/doctest_xml_compare.py
def text_compare(t1, t2):
    if not t1 and not t2:
        return True
    if t1 == "*" or t2 == "*":
        return True
    return (t1 or "").strip() == (t2 or "").strip()


def xml_compare(x1, x2, reporter=None):
    """
    Compare two XML files
    x1 must be the plain version
    x2 must be the version generated by aapt
    """
    if x1.tag != x2.tag:
        if reporter:
            reporter("Tags do not match: {} and {}".format(x1.tag, x2.tag))
        return False
    for name, value in x1.attrib.items():
        if value[0] == "@" and x2.attrib.get(name)[0] == "@":
            # Can not be sure...
            pass
        elif x2.attrib.get(name) != value:
            if reporter:
                reporter(
                    "Attributes do not match: %s=%r, %s=%r"
                    % (name, value, name, x2.attrib.get(name))
                )
            return False
    for name in x2.attrib.keys():
        if name not in x1.attrib:
            if (
                x2.tag == "application"
                and name == "{http://schemas.android.com/apk/res/android}debuggable"
            ):
                # Debug attribute might be added by aapt
                pass
            else:
                if reporter:
                    reporter("x2 has an attribute x1 is missing: %s" % name)
                return False
    if not text_compare(x1.text, x2.text):
        if reporter:
            reporter("text: {!r} != {!r}".format(x1.text, x2.text))
        return False
    if not text_compare(x1.tail, x2.tail):
        if reporter:
            reporter("tail: {!r} != {!r}".format(x1.tail, x2.tail))
        return False
    cl1 = x1.getchildren()
    cl2 = x2.getchildren()
    if len(cl1) != len(cl2):
        if reporter:
            reporter("children length differs, %i != %i" % (len(cl1), len(cl2)))
        return False
    i = 0
    for c1, c2 in zip(cl1, cl2):
        i += 1
        if not xml_compare(c1, c2, reporter=reporter):
            if reporter:
                reporter("children %i do not match: %s" % (i, c1.tag))
            return False
    return True


class AXMLTest(unittest.TestCase):

    def testAndroidManifest(self):
        filenames = [
            "data/AXML/AndroidManifest.xml",
            "data/AXML/AndroidManifest-Chinese.xml",
            "data/AXML/AndroidManifestDoubleNamespace.xml",
            "data/AXML/AndroidManifestExtraNamespace.xml",
            "data/AXML/AndroidManifest_InvalidCharsInAttribute.xml",
            "data/AXML/AndroidManifestLiapp.xml",
            "data/AXML/AndroidManifestMaskingNamespace.xml",
            "data/AXML/AndroidManifest_NamespaceInAttributeName.xml",
            "data/AXML/AndroidManifest_NamespaceInAttributeName2.xml",
            "data/AXML/AndroidManifestNonZeroStyle.xml",
            "data/AXML/AndroidManifestNullbytes.xml",
            "data/AXML/AndroidManifestTextChunksXML.xml",
            # "data/AXML/AndroidManifestUTF8Strings.xml",
            "data/AXML/AndroidManifestWithComment.xml",
            "data/AXML/AndroidManifest_WrongChunkStart.xml",
            "data/AXML/AndroidManifest-xmlns.xml",
        ]

        for filename in filenames:
            with open(os.path.join(test_dir, filename), "rb") as fd:
                ap = pyaxml.AXML.from_axml(fd.read())[0]
            self.assertIsNotNone(ap)

    def testAXMLnotAligned(self):
        filenames = [
            "data/AXML/StringBlocksNotAligned.xml",
        ]

        for filename in filenames:
            with open(os.path.join(test_dir, filename), "rb") as fp:
                ap = pyaxml.AXML.from_axml(fp.read())[0]
                ap.compute()
                sb = pyaxml.StringBlocks(proto=ap.proto.stringblocks)
                sb_s = (
                    b"".join(
                        pyaxml.StringBlock(
                            proto=elt,
                            utf8=sb.proto.hnd.flag & pyaxml.axml_pb2.UTF8_FLAG
                            == pyaxml.axml_pb2.UTF8_FLAG,
                        ).pack()
                        for elt in sb.proto.stringblocks
                    )
                    + sb.proto.pad_stringblocks
                )
            self.assertEqual(len(sb_s) % 4, 0)

    def testUnpackingRepackingXMLFormCompute(self):
        filenames = [
            "data/AXML/AndroidManifest.xml",
            "data/AXML/AndroidManifest-Chinese.xml",
            "data/AXML/AndroidManifestDoubleNamespace.xml",
            "data/AXML/AndroidManifestExtraNamespace.xml",
            "data/AXML/AndroidManifestLiapp.xml",
            "data/AXML/AndroidManifestMaskingNamespace.xml",
            "data/AXML/AndroidManifest_NamespaceInAttributeName.xml",
            "data/AXML/AndroidManifest_NamespaceInAttributeName2.xml",
            "data/AXML/AndroidManifestNonZeroStyle.xml",
            "data/AXML/AndroidManifestTextChunksXML.xml",
            # "data/AXML/AndroidManifestUTF8Strings.xml", # TODO: check if attribute exist or not
            "data/AXML/AndroidManifestWithComment.xml",
            "data/AXML/AndroidManifest_WrongChunkStart.xml",
            "data/AXML/AndroidManifest-xmlns.xml",
            "data/AXML/StringBlocksNotAligned.xml",
        ]

        for filename in filenames:
            content = ""
            with open(os.path.join(test_dir, filename), "rb") as fd:
                ap = pyaxml.AXML.from_axml(fd.read())[0]
                content = ElementTree.tostring(ap.to_xml(), encoding="utf8")
                ap.compute()
                new = ap.pack()
                ap2 = pyaxml.AXML.from_axml(new)[0]
                content2 = ElementTree.tostring(ap2.to_xml(), encoding="utf8")
                self.assertEqual(content, content2)

    def testUnpackingRepackingXMLForm(self):
        filenames = [
            "data/AXML/AndroidManifest.xml",
            "data/AXML/AndroidManifest-Chinese.xml",
            "data/AXML/AndroidManifestDoubleNamespace.xml",
            "data/AXML/AndroidManifestExtraNamespace.xml",
            "data/AXML/AndroidManifestLiapp.xml",
            "data/AXML/AndroidManifestMaskingNamespace.xml",
            "data/AXML/AndroidManifest_NamespaceInAttributeName.xml",
            "data/AXML/AndroidManifest_NamespaceInAttributeName2.xml",
            "data/AXML/AndroidManifestNonZeroStyle.xml",
            "data/AXML/AndroidManifestTextChunksXML.xml",
            # "data/AXML/AndroidManifestUTF8Strings.xml", # TODO: check if attribute exist or not
            "data/AXML/AndroidManifestWithComment.xml",
            "data/AXML/AndroidManifest_WrongChunkStart.xml",
            "data/AXML/AndroidManifest-xmlns.xml",
        ]

        for filename in filenames:
            content = ""
            with open(os.path.join(test_dir, filename), "rb") as fd:
                ap = pyaxml.AXML.from_axml(fd.read())[0]
                content = ElementTree.tostring(ap.to_xml(), encoding="utf8")
                new = ap.pack()
                ap2 = pyaxml.AXML.from_axml(new)[0]
                content2 = ElementTree.tostring(ap2.to_xml(), encoding="utf8")
                self.assertEqual(content, content2)

    def testUnpackingRepacking(self):
        filenames = [
            "data/AXML/AndroidManifest.xml",
            "data/AXML/AndroidManifest-Chinese.xml",
            "data/AXML/AndroidManifestDoubleNamespace.xml",
            "data/AXML/AndroidManifestExtraNamespace.xml",
            "data/AXML/AndroidManifest_InvalidCharsInAttribute.xml",
            "data/AXML/AndroidManifestLiapp.xml",
            "data/AXML/AndroidManifestMaskingNamespace.xml",
            "data/AXML/AndroidManifest_NamespaceInAttributeName.xml",
            "data/AXML/AndroidManifest_NamespaceInAttributeName2.xml",
            "data/AXML/AndroidManifestNonZeroStyle.xml",
            "data/AXML/AndroidManifestNullbytes.xml",
            "data/AXML/AndroidManifestTextChunksXML.xml",
            "data/AXML/AndroidManifestUTF8Strings.xml",
            "data/AXML/AndroidManifestWithComment.xml",
            "data/AXML/AndroidManifest_WrongChunkStart.xml",
            "data/AXML/AndroidManifest-xmlns.xml",
            "data/AXML/StringBlocksNotAligned.xml",
        ]

        for filename in filenames:
            with open(os.path.join(test_dir, filename), "rb") as fd:
                old = fd.read()
                ap = pyaxml.AXML.from_axml(old)[0]
                new = ap.pack()
                self.assertEqual(old, new)

    def testNonManifest(self):
        filenames = [
            "data/AXML/test.xml",
            "data/AXML/test1.xml",
            "data/AXML/test2.xml",
            "data/AXML/test3.xml",
        ]

        for filename in filenames:
            with open(os.path.join(test_dir, filename), "rb") as fp:
                ap = pyaxml.AXML.from_axml(fp.read())[0]

            self.assertEqual(ap.to_xml().tag, "LinearLayout")

    def testNonZeroStyleOffset(self):
        """
        Test if a nonzero style offset in the string section causes problems
        if the counter is 0
        """
        filename = "data/AXML/AndroidManifestNonZeroStyle.xml"

        with open(os.path.join(test_dir, filename), "rb") as f:
            ap = pyaxml.AXML.from_axml(f.read())[0].to_xml()

    def testNonTerminatedString(self):
        """
        Test if non-null terminated strings are detected.
        This sample even segfaults aapt...
        """
        filename = "data/AXML/AndroidManifest_StringNotTerminated.xml"

        # TODO
        # with open(os.path.join(test_dir, filename), "rb") as f:
        #    ap = pyaxml.AXML.from_axml(f.read())[0].to_xml()

    def testExtraNamespace(self):
        """
        Test if extra namespaces cause problems
        """
        filename = "data/AXML/AndroidManifestExtraNamespace.xml"

        with open(os.path.join(test_dir, filename), "rb") as f:
            ap = pyaxml.AXML.from_axml(f.read())[0].to_xml()

    def testTextChunksWithXML(self):
        """
        Test for Text chunks containing XML
        """
        filename = "data/AXML/AndroidManifestTextChunksXML.xml"

        # TODO
        # with open(os.path.join(test_dir, filename), "rb") as f:
        #    ap = pyaxml.AXML.from_axml(f.read())[0].to_xml()

    def testWrongFilesize(self):
        """
        Assert that files with a broken filesize are not parsed
        """
        filename = "data/AXML/AndroidManifestWrongFilesize.xml"

        try:
            with open(os.path.join(test_dir, filename), "rb") as f:
                ap = pyaxml.AXML.from_axml(f.read())[0].to_xml()
        except IndexError:
            return
        self.assertTrue(False, "should raise an error")

    def testNullbytes(self):
        """
        Assert that Strings with nullbytes are handled correctly
        """
        filename = "data/AXML/AndroidManifestNullbytes.xml"

        try:
            with open(os.path.join(test_dir, filename), "rb") as f:
                ap = pyaxml.AXML.from_axml(f.read())[0].to_xml()
        except ValueError as e:
            self.assertEqual(
                str(e),
                "All strings must be XML compatible: Unicode or ASCII, no NULL bytes or control characters",
            )
            return
        self.assertTrue(False, "should raise an error")

    def testMaskingNamespace(self):
        """
        Assert that Namespaces which are used in a tag and the tag is closed
        are actually correctly parsed.
        """
        filename = "data/AXML/AndroidManifestMaskingNamespace.xml"

        with open(os.path.join(test_dir, filename), "rb") as f:
            ap = pyaxml.AXML.from_axml(f.read())[0].to_xml()

    def testDoubleNamespace(self):
        """
        Test if weird namespace constelations cause problems
        """
        filename = "data/AXML/AndroidManifestDoubleNamespace.xml"

        with open(os.path.join(test_dir, filename), "rb") as f:
            ap = pyaxml.AXML.from_axml(f.read())[0].to_xml()

    def testPackers(self):
        """
        Assert that Packed files are read
        """
        filename = "data/AXML/AndroidManifestLiapp.xml"

        with open(os.path.join(test_dir, filename), "rb") as f:
            ap = pyaxml.AXML.from_axml(f.read())[0].to_xml()


if __name__ == "__main__":
    unittest.main()



import sys
from struct import pack, unpack
import logging
from typing import Union

try:
    from pyaxml.proto import axml_pb2
except ImportError:
    print("proto is not build")
    sys.exit(1)
from pyaxml.stringblocks import AXMLHeader
from pyaxml.arsctype import ARSCResStringPoolRef


##############################################################################
#
#                              XML ELEMENTS
#
##############################################################################


class AXMLHeaderResXml(AXMLHeader):
    """AXMLHeaderResXml class to build an header of RES_XML"""

    def __init__(self, type_=0, size=0, proto: axml_pb2.AXMLHeader = None):
        """Initialize header of Res_XML

        Args:
            type (int, optional): define the type. Defaults to 0.
            size (int, optional): define the size of whole element. Defaults to 0.
            proto (axml_pb2.AXMLHeader, optional): define RES_XML header by a protobuff. Defaults to None.
        """
        if proto is None:
            super().__init__(type_, size + 8, proto)
            self.proto.header_size = 16
        else:
            self.proto = proto


class AXMLHeaderStartElement(AXMLHeaderResXml):
    """AXMLHeaderStartElement class to build an header of Start element"""

    def __init__(self, size: int):
        """initialize START_ELEMENT element

        Args:
            size (int): size of START_ELEMENT and its header
        """
        super().__init__(axml_pb2.RES_XML_START_ELEMENT_TYPE, size)


class AXMLHeaderEndElement(AXMLHeaderResXml):
    """AXMLHeaderEndElement class to build an header of End element"""

    def __init__(self, size: int):
        """initialize END_ELEMENT element

        Args:
            size (int): size of END_ELEMENT and its header
        """
        super().__init__(axml_pb2.RES_XML_END_ELEMENT_TYPE, size)


class AXMLHeaderStartNamespace(AXMLHeaderResXml):
    """AXMLHeaderStartNamespace class to build an header of Start namespace"""

    def __init__(self, size: int):
        """initialize START_NAMESPACE element

        Args:
            size (int): size of START_NAMESPACE and its header
        """
        super().__init__(axml_pb2.RES_XML_START_NAMESPACE_TYPE, size)


class AXMLHeaderEndNamespace(AXMLHeaderResXml):
    """AXMLHeaderEndNamespace class to build an header of End namespace"""

    def __init__(self, size: int):
        """initialize END_NAMESPACE element

        Args:
            size (int): size of END_NAMESPACE and its header
        """
        super().__init__(axml_pb2.RES_XML_END_NAMESPACE_TYPE, size)


class ClassicalResXml:
    """RES_XML class to build RES_XML element"""

    proto: axml_pb2.ResXML = None

    def __init__(
        self,
        linenumber: int = 0,
        comment: int = 0xFFFFFFFF,
        proto: axml_pb2.ResXML = None,
    ):
        """initialize RES_XML element

        Args:
            lineNumber (int, optional): _description_. Defaults to 0.
            Comment (int, optional): _description_. Defaults to 0xffffffff.
            proto (axml_pb2.ResXML, optional): define RES_XML by a protobuff. Defaults to None.
        """
        if proto is None:
            self.proto.generic.lineNumber = linenumber
            self.proto.generic.Comment = comment
        else:
            self.proto.generic.CopyFrom(proto)

    @property
    def content(self) -> bytes:
        """get content of classical res XML

        Returns:
            bytes: content packed
        """
        return pack("<LL", self.proto.generic.lineNumber, self.proto.generic.Comment)

    def compute(self):
        """Compute all fields to have all RES_XML elements"""

    def pack(self) -> bytes:
        """pack the RES_XML element

        Returns:
            bytes: return the RES_XML element packed
        """
        return self.content

    @staticmethod
    def from_axml(buff: bytes, class_xml=None):
        """Convert ClassicalResXml buffer to ClassicalResXml object

        Args:
            buff (bytes): buffer contain ClassicalResXml object
            class_xml (pyaxml.ClassicalResXml): _description_

        Returns:
            tuple[pyaxml.ClassicalResXml, bytes]: return ClassicalResXml element and buffer offset at the end of the reading
        """
        if class_xml is None:
            raise ValueError("class_xml argument not set")
        class_xml.proto.generic.lineNumber, class_xml.proto.generic.Comment = unpack(
            "<LL", buff[:8]
        )
        return class_xml, buff[8:]


class ResXmlStartElement(ClassicalResXml):
    """ResXmlStartElement"""

    def __init__(
        self,
        namespaceuri: int = 0xFFFFFFFF,
        name: int = 0xFFFFFFFF,
        attributes: Union[list, None] = None,
        styleAttribute: int = -1,
        classAttribute: int = -1,
        lineNumber: int = 0,
        Comment: int = 0xFFFFFFFF,
        at_start=0x14,
        at_size=0x14,
        proto: axml_pb2.ResXMLStartElement = None,
    ):
        """_summary_

        Args:
            namespaceuri (int, optional): _description_. Defaults to 0xffffffff.
            name (int, optional): _description_. Defaults to 0xffffffff.
            attributes (list, optional): _description_. Defaults to [].
            styleAttribute (int, optional): _description_. Defaults to -1.
            classAttribute (int, optional): _description_. Defaults to -1.
            lineNumber (int, optional): _description_. Defaults to 0.
            Comment (int, optional): _description_. Defaults to 0xffffffff.
            proto (axml_pb2.ResXMLStartElement, optional): _description_. Defaults to None.
        """
        if attributes is None:
            attributes = []
        if proto is None:
            self.proto = axml_pb2.ResXMLStartElement()
            super().__init__(lineNumber, Comment)
            self.proto.namespaceURI = namespaceuri
            self.proto.name = name
            self.proto.attributes.extend(attributes)
            self.proto.styleAttribute = styleAttribute
            self.proto.classAttribute = classAttribute
            self.proto.at_start = at_start
            self.proto.at_size = at_size
        else:
            self.proto = proto
            super().__init__(proto=proto.generic)

    def compute(self):
        """Compute all fields to have all ResXmlStartElement elements"""
        self.proto.len_attributes = len(self.proto.attributes)
        super().compute()

    @property
    def content(self) -> bytes:
        """_summary_

        Returns:
            bytes: _description_
        """
        return (
            super().content
            + pack(
                "<LLhhLhh",
                self.proto.namespaceURI,
                self.proto.name,
                self.proto.at_start,
                self.proto.at_size,
                self.proto.len_attributes,
                self.proto.styleAttribute,
                self.proto.classAttribute,
            )
            + b"".join(Attribute(proto=a).pack() for a in self.proto.attributes)
        )

    @staticmethod
    def from_axml(buff: bytes, class_xml=None, proto=None):
        """Convert ResXmlStartElement buffer to ResXmlStartElement object

        Args:
            buff (bytes): buffer contain ResXmlStartElement object

        Returns:
            tuple[pyaxml.ResXmlStartElement, bytes]: return ResXmlStartElement element and buffer offset at the end of the reading
        """
        if class_xml is None:
            class_xml = ResXmlStartElement()
        if proto:
            class_xml.proto = proto
        class_xml, buff = ClassicalResXml.from_axml(buff, class_xml=class_xml)

        (
            class_xml.proto.namespaceURI,
            class_xml.proto.name,
            class_xml.proto.at_start,
            class_xml.proto.at_size,
            class_xml.proto.len_attributes,
            class_xml.proto.styleAttribute,
            class_xml.proto.classAttribute,
        ) = unpack("<LLhhLhh", buff[:20])
        buff = buff[20:]
        # TODO len_attributes contain id_attribute and count attribute
        for _ in range(class_xml.proto.len_attributes & 0xFFFF):
            padding = class_xml.proto.at_size - 0x14
            attr, buff = Attribute.from_axml(buff, padding)
            class_xml.proto.attributes.append(attr.proto)

        return class_xml, buff


class ResXmlEndElement(ClassicalResXml):
    """ResXmlEndElement"""

    def __init__(
        self,
        namespaceURI: int = 0xFFFFFFFF,
        name: int = 0xFFFFFFFF,
        lineNumber: int = 0,
        Comment: int = 0xFFFFFFFF,
        proto: axml_pb2.ResXMLEndElement = None,
    ):
        """_summary_

        Args:
            namespaceURI (int, optional): _description_. Defaults to 0xffffffff.
            name (int, optional): _description_. Defaults to 0xffffffff.
            lineNumber (int, optional): _description_. Defaults to 0.
            Comment (int, optional): _description_. Defaults to 0xffffffff.
            proto (axml_pb2.ResXMLEndElement, optional): _description_. Defaults to None.
        """
        if proto is None:
            self.proto = axml_pb2.ResXMLEndElement()
            super().__init__(lineNumber, Comment)
            self.proto.namespaceURI = namespaceURI
            self.proto.name = name
        else:
            self.proto = proto
            super().__init__(proto=proto.generic)

    @property
    def content(self) -> bytes:
        """_summary_

        Returns:
            bytes: _description_
        """
        return super().content + pack("<LL", self.proto.namespaceURI, self.proto.name)

    @staticmethod
    def from_axml(buff, class_xml=None, proto=None):
        """Convert ResXmlEndElement buffer to ResXmlEndElement object

        Args:
            buff (bytes): buffer contain ResXmlEndElement object

        Returns:
            tuple[pyaxml.ResXmlEndElement, bytes]: return ResXmlEndElement element and buffer offset at the end of the reading
        """
        if class_xml is None:
            class_xml = ResXmlEndElement()
        if proto:
            class_xml.proto = proto
        class_xml, buff = ClassicalResXml.from_axml(buff, class_xml=class_xml)
        class_xml.proto.namespaceURI, class_xml.proto.name = unpack("<LL", buff[:8])
        return class_xml, buff[8:]


class ResXmlStartNamespace(ClassicalResXml):
    """ResXmlStartNamespace"""

    def __init__(
        self,
        prefix: int = 0xFFFFFFFF,
        uri: int = 0xFFFFFFFF,
        lineNumber: int = 0,
        Comment: int = 0xFFFFFFFF,
        proto: axml_pb2.ResXMLStartNamespace = None,
    ):
        """_summary_

        Args:
            prefix (int, optional): _description_. Defaults to 0xffffffff.
            uri (int, optional): _description_. Defaults to 0xffffffff.
            lineNumber (int, optional): _description_. Defaults to 0.
            Comment (int, optional): _description_. Defaults to 0xffffffff.
            proto (axml_pb2.ResXMLStartNamespace, optional): _description_. Defaults to None.
        """
        if proto is None:
            self.proto = axml_pb2.ResXMLStartNamespace()
            super().__init__(lineNumber, Comment)
            self.proto.prefix = prefix
            self.proto.uri = uri
        else:
            self.proto = proto
            super().__init__(proto=proto.generic)

    @property
    def content(self) -> bytes:
        """_summary_

        Returns:
            bytes: _description_
        """
        return super().content + pack("<LL", self.proto.prefix, self.proto.uri)

    @staticmethod
    def from_axml(buff: bytes, class_xml=None, proto=None):
        """Convert ResXmlStartNamespace buffer to ResXmlStartNamespace object

        Args:
            buff (bytes): buffer contain ResXmlStartNamespace object

        Returns:
            tuple[pyaxml.ResXmlStartNamespace, bytes]: return ResXmlStartNamespace element and buffer offset at the end of the reading
        """
        if class_xml is None:
            class_xml = ResXmlStartNamespace()
        if proto:
            class_xml.proto = proto
        class_xml, buff = ClassicalResXml.from_axml(buff, class_xml=class_xml)
        class_xml.proto.prefix, class_xml.proto.uri = unpack("<LL", buff[:8])
        return class_xml, buff[8:]


class ResXmlEndNamespace(ClassicalResXml):
    """ResXmlEndNamespace"""

    def __init__(
        self,
        prefix: int = 0xFFFFFFFF,
        uri: int = 0xFFFFFFFF,
        lineNumber: int = 0,
        Comment: int = 0xFFFFFFFF,
        proto: axml_pb2.ResXMLEndNamespace = None,
    ):
        """_summary_

        Args:
            prefix (int, optional): _description_. Defaults to 0xffffffff.
            uri (int, optional): _description_. Defaults to 0xffffffff.
            lineNumber (int, optional): _description_. Defaults to 0.
            Comment (int, optional): _description_. Defaults to 0xffffffff.
            proto (axml_pb2.ResXMLEndNamespace, optional): _description_. Defaults to None.
        """
        if proto is None:
            self.proto = axml_pb2.ResXMLEndNamespace()
            super().__init__(lineNumber, Comment)
            self.proto.prefix = prefix
            self.proto.uri = uri
        else:
            self.proto = proto
            super().__init__(proto=proto.generic)

    @property
    def content(self):
        """_summary_

        Returns:
            _type_: _description_
        """
        return super().content + pack("<LL", self.proto.prefix, self.proto.uri)

    @staticmethod
    def from_axml(buff: bytes, class_xml=None, proto=None):
        """Convert ResXmlEndNamespace buffer to ResXmlEndNamespace object

        Args:
            buff (bytes): buffer contain ResXmlEndNamespace object

        Returns:
            tuple[pyaxml.ResXmlEndNamespace, bytes]: return ResXmlEndNamespace element and buffer offset at the end of the reading
        """
        if class_xml is None:
            class_xml = ResXmlEndNamespace()
        if proto:
            class_xml.proto = proto
        class_xml, buff = ClassicalResXml.from_axml(buff, class_xml=class_xml)
        class_xml.proto.prefix, class_xml.proto.uri = unpack("<LL", buff[:8])
        return class_xml, buff[8:]


class Attribute:
    """Attribute"""

    def __init__(
        self,
        namespaceuri: int = 0xFFFFFFFF,
        name: int = 0xFFFFFFFF,
        value: int = 0xFFFFFFFF,
        type_: int = 0xFFFFFFFF,
        data: int = 0xFFFFFFFF,
        proto: axml_pb2.Attribute = None,
    ):
        """Initialize an Attribute element from protobuff or parameters

        Args:
            namespaceURI (int, optional): namespace of Attribute. Defaults to 0xffffffff.
            name (int, optional): name of Attribute. Defaults to 0xffffffff.
            value (int, optional): value of the attribute. Defaults to 0xffffffff.
            type_ (int, optional): type of attribute. Defaults to 0xffffffff.
            data (int, optional): data of the parameter in function of the type it could be store in value or data. Defaults to 0xffffffff.
            proto (axml_pb2.Attribute, optional): protobuff of Attribute content. Defaults to None.
        """
        if proto is None:
            self.proto = axml_pb2.Attribute()
            self.proto.namespaceURI = namespaceuri
            self.proto.name = name
            self.proto.value = value
            self.proto.type = type_
            self.proto.data = data
        else:
            self.proto = proto

    def pack(self) -> bytes:
        """pack the Attribute element

        Returns:
            bytes: return the Attribute element packed
        """
        return (
            pack(
                "<LLLLL",
                self.proto.namespaceURI,
                self.proto.name,
                self.proto.value,
                self.proto.type,
                self.proto.data,
            )
            + self.proto.padding
        )

    @staticmethod
    def from_axml(buff, padding=0, proto=None):
        """Convert Attribute buffer to Attribute object

        Args:
            buff (bytes): buffer contain Attribute object

        Returns:
            tuple[pyaxml.Attribute, bytes]: return Attribute element and buffer offset at the end of the reading
        """
        attr = Attribute()
        if proto:
            attr.proto = proto
        (
            attr.proto.namespaceURI,
            attr.proto.name,
            attr.proto.value,
            attr.proto.type,
            attr.proto.data,
        ) = unpack("<LLLLL", buff[:20])
        attr.proto.padding = buff[20 : 20 + padding]
        return attr, buff[20 + padding :]


class ResXmlCDATA(ClassicalResXml):
    """ResXmlCDATA"""

    def __init__(
        self,
        proto: axml_pb2.ResXMLCDATA = None,
        lineNumber: int = 0,
        Comment: int = 0xFFFFFFFF,
    ):
        """Initialize an ResXmlCDATA element from protobuff or parameters

        Args:
            proto (axml_pb2.ResXMLCDATA, optional): _description_. Defaults to None.
        """
        if proto is None:
            self.proto = axml_pb2.ResXMLCDATA()
            super().__init__(lineNumber, Comment)
        else:
            self.proto = proto

    @property
    def content(self):
        """_summary_

        Returns:
            _type_: _description_
        """
        return (
            super().content
            + pack("<L", self.proto.name)
            + ARSCResStringPoolRef(proto=self.proto.res).pack()
        )

    @staticmethod
    def from_axml(buff, class_xml=None, proto=None):
        """Convert ResXmlCDATA buffer to ResXmlCDATA object

        Args:
            buff (bytes): buffer contain ResXmlCDATA object

        Returns:
            tuple[pyaxml.ResXmlCDATA, bytes]: return ResXmlCDATA element and buffer offset at the end of the reading
        """
        class_xml = ResXmlCDATA()
        if proto:
            class_xml.proto = proto
        class_xml, buff = ClassicalResXml.from_axml(buff, class_xml=class_xml)
        class_xml.proto.name = unpack("<L", buff[:4])[0]
        _, buff = ARSCResStringPoolRef.from_axml(buff[4:], proto=class_xml.proto.res)
        return class_xml, buff


class ResourceXML:
    """ResourceXML"""

    def __init__(self, proto: axml_pb2.ResourceXML = None) -> None:
        """Initialize ResourceXML

        Args:
            proto (axml_pb2.ResourceXML, optional): protobuff of ResourceXML content. Defaults to None.
        """
        if proto:
            self.proto = proto
        else:
            self.proto = axml_pb2.ResourceXML()

    def pack(self) -> bytes:
        """pack the ResourceXML element

        Returns:
            bytes: return the ResourceXML element packed
        """
        buf = b""
        for elt in self.proto.elts:
            header = AXMLHeader(proto=elt.header).pack()
            if elt.HasField("start_elt"):
                buf += header + ResXmlStartElement(proto=elt.start_elt).pack()
            elif elt.HasField("end_elt"):
                buf += header + ResXmlEndElement(proto=elt.end_elt).pack()
            elif elt.HasField("start_ns"):
                buf += header + ResXmlStartNamespace(proto=elt.start_ns).pack()
            elif elt.HasField("end_ns"):
                buf += header + ResXmlEndNamespace(proto=elt.end_ns).pack()
            elif elt.HasField("cdata"):
                buf += header + ResXmlCDATA(proto=elt.cdata).pack()
        return buf

    @staticmethod
    def from_axml(buff: bytes, proto=None):
        """Convert ResourceXML buffer to ResourceXML object

        Args:
            buff (bytes): buffer contain ResourceXML object

        Returns:
            tuple[pyaxml.ResourceXML, bytes]: return ResourceXML element and buffer offset at the end of the reading
        """
        xml = ResourceXML()
        if proto:
            xml.proto = proto
        while len(buff) > 0:
            elt = axml_pb2.XMLElement()
            hnd, buff = AXMLHeader.from_axml(buff, proto=elt.header)
            if hnd.proto.type == axml_pb2.RES_XML_START_ELEMENT_TYPE:
                _, buff2 = ResXmlStartElement.from_axml(buff, proto=elt.start_elt)
            elif hnd.proto.type == axml_pb2.RES_XML_END_ELEMENT_TYPE:
                _, buff2 = ResXmlEndElement.from_axml(buff, proto=elt.end_elt)
            elif hnd.proto.type == axml_pb2.RES_XML_START_NAMESPACE_TYPE:
                _, buff2 = ResXmlStartNamespace.from_axml(buff, proto=elt.start_ns)
            elif hnd.proto.type == axml_pb2.RES_XML_END_NAMESPACE_TYPE:
                _, buff2 = ResXmlEndNamespace.from_axml(buff, proto=elt.end_ns)
            elif hnd.proto.type == axml_pb2.RES_XML_CDATA_TYPE:
                _, buff2 = ResXmlCDATA.from_axml(buff, proto=elt.cdata)
            else:
                raise ValueError("RES_XML element header incorrect type")
            buff = buff[hnd.proto.size - hnd.proto.header_size + 8 :]
            if buff != buff2:
                logging.warning(
                    "Size of resourceXML element incorrect from type %s the size difference is %s",
                    hnd.proto.type,
                    len(buff2) - len(buff),
                )
            xml.proto.elts.append(elt)
        return xml, buff



import os
from typing import Union
from xml.dom import minidom

_public_res: Union[dict[str, dict], None] = None
# copy the newest sdk/platforms/android-?/data/res/values/public.xml here
if _public_res is None:
    _public_res = {}
    root = os.path.dirname(os.path.realpath(__file__))
    xmlfile = os.path.join(root, "public.xml")
    if os.path.isfile(xmlfile):
        with open(xmlfile, "r", encoding="utf-8") as fp:
            _xml = minidom.parseString(fp.read())
            for element in _xml.getElementsByTagName("public"):
                _type = element.getAttribute("type")
                _name = element.getAttribute("name")
                if element.getAttribute("id") == "":
                    ID_ = None
                else:
                    ID_ = int(element.getAttribute("id"), 16)
                if _type not in _public_res:
                    _public_res[_type] = {}
                _public_res[_type][_name] = ID_
    else:
        raise ValueError("need to copy the sdk/platforms/android-?/data/res/values/public.xml here")

SYSTEM_RESOURCES = {
    "attributes": {
        "forward": dict(_public_res["attr"].items()),
        "inverse": {v: k for k, v in _public_res["attr"].items()},
    },
    "styles": {
        "forward": dict(_public_res["style"].items()),
        "inverse": {v: k for k, v in _public_res["style"].items()},
    },
}


# if __name__ == '__main__':
#    import json
#    _resources = None
#    if _resources is None:
#        root = os.path.dirname(os.path.realpath(__file__))
#        resfile = os.path.join(root, "public.json")
#
#        if os.path.isfile(resfile):
#            with open(resfile, "r") as fp:
#                _resources = json.load(fp)
#        else:
#            # TODO raise error instead?
#            _resources = {}
#    for _type in set([] + list(_public_res.keys()) + list(_resources.keys())):
#        for k in set([] + list(_public_res.get(_type, {}).keys())
#                     + list(_resources.get(_type, {}).keys())):
#            a,b = _public_res.get(_type, {}).get(k), \
#                  _resources.get(_type, {}).get(k),
#            if a != b:
#                print(k, a,b)
#    print(None)



from struct import pack, unpack
import sys
from typing import Union

try:
    from pyaxml.proto import axml_pb2
except ImportError:
    print("proto is not build")
    sys.exit(1)

from pyaxml.stringblocks import StringBlocks
from pyaxml.header import AXMLHeaderRESTABLE, TypedValue, AXMLHeader, AXML_HEADER_SIZE
from pyaxml.arsctype import (
    ARSCResTableConfig,
    ARSCResTableEntry,
    ARSCResTypeSpec,
    ARSCResTypeType,
)

##############################################################################
#
#                              ARSC OBJECT
#
##############################################################################


class ARSC:
    """ARSC"""

    def __init__(self, proto: axml_pb2.ARSC = None):
        """_summary_

        Args:
            proto (axml_pb2.AXML, optional): _description_. Defaults to None.
        """
        if proto:
            self.proto = proto
        else:
            self.proto = axml_pb2.ARSC()

    @property
    def get_proto(self):
        """_summary_

        Returns:
            _type_: _description_
        """
        return self.proto

    def compute(self, recursive=True):
        """Compute all fields to have all AXML elements"""
        if recursive:
            s = StringBlocks(proto=self.proto.stringblocks)
            s.compute(update_size=True)
            self.proto.stringblocks.CopyFrom(s.proto)
            for _, restablespackage in enumerate(
                self.proto.restablespackage
            ):  # range(len(self.proto.restablespackage)):
                package = AXMLResTablePackage(proto=restablespackage)
                package.compute(recursive=recursive)
                restablespackage.CopyFrom(package.proto)

        self.proto.header_res.hnd.size = len(self.pack())
        self.proto.header_res.package_count = len(self.proto.restablespackage)

    def pack(self) -> bytes:
        """pack the AXML element

        Returns:
            bytes: return the AXML element packed
        """
        sb_pack = StringBlocks(proto=self.proto.stringblocks).pack()
        header_res = AXMLHeaderRESTABLE(proto=self.proto.header_res).pack()
        packages = b""
        for package in self.proto.restablespackage:
            packages += AXMLResTablePackage(proto=package).pack()
        return header_res + sb_pack + packages

    @staticmethod
    def from_axml(buff: bytes, proto=None):
        """read an AXML ARSC file to convert in pyaxml.ARSC object

        Args:
            buff (bytes): the string buffer to read

        Returns:
            ARSC: the ARSC object
            str: index of the reading string, if it's a valid file it should be the EOF.
        """
        arsc = ARSC()
        if proto:
            arsc.proto = proto
        header, buff = AXMLHeaderRESTABLE.from_axml(buff, proto=arsc.proto.header_res)
        _, buff = StringBlocks.from_axml(buff, proto=arsc.proto.stringblocks)
        for _ in range(header.proto.package_count):
            restablepackage, buff = AXMLResTablePackage.from_axml(buff)
            arsc.proto.restablespackage.append(restablepackage.proto)

        return arsc, buff

    @staticmethod
    def convert_id(type_id: int, id_: int) -> int:
        """convert id and type to a general id used on AXML file to refer to resources

        Args:
            type_id (int): type of value (xml, string, etc.)
            id (int): index from the list of type value

        Returns:
            int: the id built
        """
        return 0x7F000000 | (type_id & 0xFF) << 16 | type_id & 0xFF00 | id_

    @staticmethod
    def get_type_from_id(id_: int) -> int:
        """get type from id

        Args:
            id_ (int): full id

        Returns:
            int: type
        """
        return (id_ & 0xFF0000) >> 16 | id_ & 0xFF00

    @staticmethod
    def get_local_id_from_id(id_: int) -> int:
        """get local id from full id

        Args:
            id_ (int): full id

        Returns:
            int: id
        """
        return id_ & 0xFF

    def get_id_public(self, package: str, type_: str, name: str) -> Union[tuple[int, int], None]:
        """get ID public from resource

        Args:
            package (str): package name
            type_ (str): type string
            name (str): name of value

        Returns:
            tuple[int, int] | None: return the id and the stringblock index of the value
        """
        for p in self.proto.restablespackage:
            n2 = p.name.split("\x00")[0]
            if n2 == package:
                for r in p.restypes:
                    if r.hnd.type == axml_pb2.ResType.RES_TABLE_TYPE_TYPE:
                        for id_, table in enumerate(r.typetype.tables):
                            index = table.index
                            n = StringBlocks(proto=p.key_sp_string).decode_str(index)
                            t = StringBlocks(proto=p.type_sp_string).decode_str(r.typetype.id - 1)
                            if n == name and t == type_:
                                return (
                                    ARSC.convert_id(r.typetype.id, id_),
                                    table.key.data,
                                )
        return None

    def get_id(
        self, package: str, rid: int
    ) -> tuple[Union[str, None], Union[str, None], Union[int, None]]:
        """get id

        Args:
            package (str): target package
            rid (int): target rid

        Returns:
            tuple[str, str, int]: type, key and rid
        """
        for p in self.proto.restablespackage:
            n2 = p.name.split("\x00")[0]
            if n2 == package:
                for r in p.restypes:
                    if r.hnd.type != axml_pb2.ResType.RES_TABLE_TYPE_TYPE:
                        continue
                    if r.typetype.id != ARSC.get_type_from_id(rid):
                        continue
                    local_id = ARSC.get_local_id_from_id(rid)
                    if local_id >= len(r.typetype.tables):
                        continue
                    table = r.typetype.tables[local_id]
                    n = StringBlocks(proto=p.key_sp_string).decode_str(table.index)
                    t = StringBlocks(proto=p.type_sp_string).decode_str(r.typetype.id - 1)
                    return t, n, rid
        return None, None, None

    def get_resource_xml_name(
        self, r_id: int, package: Union[str, None] = None
    ) -> Union[str, None]:
        """get resource xml name

        Args:
            r_id (int): rid target
            package (Union[str, None], optional): target package. Defaults to None.

        Returns:
            str: xml name from resource
        """
        if package:
            resource, name, i_id = self.get_id(package, r_id)
            if not i_id:
                return None
            return "@{}/{}".format(resource, name)

        for p in self.get_packages():
            r, n, i_id = self.get_id(p, r_id)
            if i_id:
                # found the resource in this package
                package = p
                resource = r
                name = n
                break
        if not package:
            return None
        return "@{}:{}/{}".format(package, resource, name)

    def get_value(self, package: str, rid: int) -> str:
        """get value of resource

        Args:
            package (str): target package
            rid (int): target rid

        Raises:
            ValueError: if rid could not be decoded

        Returns:
            str: value decoded
        """
        general_st = StringBlocks(proto=self.proto.stringblocks)
        for p in self.proto.restablespackage:
            n2 = p.name.split("\x00")[0]
            if n2 == package:
                for r in p.restypes:
                    if r.hnd.type != axml_pb2.ResType.RES_TABLE_TYPE_TYPE:
                        continue
                    if r.typetype.id != ARSC.get_type_from_id(rid):
                        continue
                    local_id = ARSC.get_local_id_from_id(rid)
                    if local_id >= len(r.typetype.tables):
                        continue
                    table = r.typetype.tables[local_id]
                    if table.key.data_type == 3:
                        return general_st.decode_str(table.key.data)
        raise ValueError(f"no value could be decoded for rid: {rid}")

    def add_id_public(self, package: str, type_: str, name: str, path: str) -> Union[int, None]:
        """Add a new id in public string on ARSC resource

        Args:
            package (str): target package
            type_ (str): type of public element
            name (str): name of public element
            path (str): value of element

        Returns:
            int | None: return the id injected
        """
        for p in self.proto.restablespackage:
            n2 = p.name.split("\x00")[0]
            if n2 == package:
                s = StringBlocks(proto=self.proto.stringblocks)
                id_path = s.get(path)
                s.compute()
                self.proto.stringblocks.CopyFrom(s.proto)
                res = AXMLResTablePackage(proto=p)
                ret = res.add_id_public(type_, name, id_path)
                res.compute()
                return ret
        return None

    def list_packages(self) -> str:
        """print all package content"""
        ret = ""
        general_st = StringBlocks(proto=self.proto.stringblocks)
        for package in self.proto.restablespackage:
            # ret += f"{package.name}\n"
            for r in package.restypes:
                if r.hnd.type == axml_pb2.ResType.RES_TABLE_TYPE_TYPE:
                    for id_, table in enumerate(r.typetype.tables):
                        if not table.present:
                            continue
                        index = table.index
                        name = StringBlocks(proto=package.key_sp_string).decode_str(index)
                        type_ = StringBlocks(proto=package.type_sp_string).decode_str(
                            r.typetype.id - 1
                        )
                        identifiant = ARSC.convert_id(r.typetype.id, id_)
                        if table.key.data_type == 3:
                            data = general_st.decode_str(table.key.data)
                        else:
                            data = hex(table.key.data)
                        data_size = table.key.size
                        ret += f'<public type="{type_}" name="{name}" id="{hex(identifiant)}" data="{data}" data_size={data_size}/>\n'
        return ret

    def get_packages(self) -> list[str]:
        """get in a list all package name of the resources.

        Returns:
            list[str]: the list of all packages
        """
        packages = []
        for p in self.proto.restablespackage:
            packages.append(p.name.split("\x00")[0])
        return packages

    def get_type_configs(
        self, package_name: str, type_name: Union[str, None] = None
    ) -> "dict[str, axml_pb2.ARSCResTableConfig]":
        """get type configuration

        Args:
            package_name (str): _description_
            type_name (str, optional): _description_. Defaults to None.

        Returns:
            dict[str, axml_pb2.ARSCResTableConfig]: type configs
        """
        if package_name is None:
            package_name = self.get_packages()[0]
        result: dict[str, axml_pb2.ARSCResTableConfig] = {}

        for restype in self.proto.restablespackage[0].restypes:
            if restype.hnd.type == axml_pb2.ResType.RES_TABLE_TYPE_TYPE:
                t_name = StringBlocks(
                    proto=self.proto.restablespackage[0].type_sp_string
                ).decode_str(restype.typetype.id - 1)
                if type_name is None or t_name == type_name:
                    result[t_name].append(restype.typetype.config)

        return result

    def get_res_configs(
        self, rid: int, config: Union[str, None] = None, fallback: bool = True
    ) -> "list[tuple[ARSCResTableConfig, ARSCResTableEntry]]":
        """get ressource configuration

        Args:
            rid (int): rid value
            config (_type_, optional): config value. Defaults to None.
            fallback (bool, optional): fallback value. Defaults to True.

        Raises:
            ValueError: if rid is not set
            ValueError: if rid is not the correct type (int)

        Returns:
            list[tuple[ARSCResTableConfig, ARSCResTableEntry]]: the list of ressource configuration
        """

        if not rid:
            raise ValueError("'rid' should be set")
        if not isinstance(rid, int):
            raise ValueError("'rid' must be an int")
        res: list[tuple[ARSCResTableConfig, ARSCResTableEntry]] = []
        p = self.proto.restablespackage[0]
        for r in p.restypes:
            if r.hnd.type != axml_pb2.ResType.RES_TABLE_TYPE_TYPE:
                continue
            if r.typetype.id != ARSC.get_type_from_id(rid):
                continue
            local_id = ARSC.get_local_id_from_id(rid)
            if local_id >= len(r.typetype.tables):
                continue
            table = r.typetype.tables[local_id]
            res.append(
                (
                    ARSCResTableConfig(proto=r.typetype.config),
                    ARSCResTableEntry(proto=table),
                )
            )
        return res

    def get_resolved_res_configs(self, rid: int, config=None) -> list[tuple]:
        """
        Return a list of resolved resource IDs with their corresponding configuration.
        It has a similar return type as :meth:`get_res_configs` but also handles complex entries
        and references.
        Also instead of returning :class:`ARSCResTableEntry` in the tuple, the actual values are resolved.

        This is the preferred way of resolving resource IDs to their resources.

        :param int rid: the numerical ID of the resource
        :param ARSCTableResConfig config: the desired configuration or None to retrieve all
        :return: A list of tuples of (ARSCResTableConfig, str)
        """
        resolver = ARSC.ResourceResolver(self, config)
        return resolver.resolve(rid)

    class ResourceResolver:
        """
        Resolves resources by ID and configuration.
        This resolver deals with complex resources as well as with references.
        """

        def __init__(self, android_resources, config=None) -> None:
            """
            :param ARSCParser android_resources: A resource parser
            :param ARSCResTableConfig config: The desired configuration or None to resolve all.
            """
            self.resources = android_resources
            self.wanted_config = config

        def resolve(self, res_id: int) -> list:
            """
            the given ID into the Resource and returns a list of matching resources.

            :param int res_id: numerical ID of the resource
            :return: a list of tuples of (ARSCResTableConfig, str)
            """
            result: "list[Union[str, tuple[ARSCResTableConfig, list], None]]" = []
            self._resolve_into_result(result, res_id, self.wanted_config)
            return result

        def _resolve_into_result(
            self,
            result: "list[Union[str, tuple[ARSCResTableConfig, list], None]]",
            res_id,
            config,
        ) -> None:
            # First: Get all candidates
            configs: "list[tuple[ARSCResTableConfig, ARSCResTableEntry]]" = (
                self.resources.get_res_configs(res_id, config)
            )

            for local_config, ate in configs:
                # deconstruct them and check if more candidates are generated
                self.put_ate_value(result, ate, local_config)

        def put_ate_value(
            self,
            result: "list[Union[str, tuple[ARSCResTableConfig, list], None]]",
            ate: axml_pb2.ARSCResTableEntry,
            config: "Union[ARSCResTableConfig, None]",
        ) -> None:
            """
            Put a ResTableEntry into the list of results
            :param list result: results array
            :param ARSCResTableEntry ate:
            :param ARSCResTableConfig config:
            :return:
            """
            if ate.is_complex():
                complex_array: "list[Union[str, tuple[ARSCResTableConfig, list], None]]" = []
                result.append((config, complex_array))
                for _, item in ate.item.items:
                    self.put_item_value(complex_array, item, config, ate, complex_=True)
            else:
                self.put_item_value(result, ate.key, config, ate, complex_=False)

        def put_item_value(
            self,
            result: list[Union[str, tuple[ARSCResTableConfig, list], None]],
            item,
            config,
            parent,
            complex_: bool,
        ) -> None:
            """
            Put the tuple (ARSCResTableConfig, resolved string) into the result set

            :param list[str | None] result: the result set
            :param ARSCResStringPoolRef item:
            :param ARSCResTableConfig config:
            :param ARSCResTableEntry parent: the originating entry
            :param bool complex_: True if the originating :class:`ARSCResTableEntry` was complex
            :return:
            """
            if item.is_reference():
                res_id = item.proto.data
                if res_id:
                    # Infinite loop detection:
                    # TODO should this stay here or should be detect the loop much earlier?
                    # if res_id == parent.mResId: # mResId == AXMLResTablePackage.id
                    #    logger.warning("Infinite loop detected at resource item {}. It references itself!".format(parent))
                    #    return

                    self._resolve_into_result(result, item.proto.data, self.wanted_config)
            else:
                if complex_:
                    if item.proto.data_type == TypedValue.TYPE_STRING:
                        v = str(item.proto.value)  # TODO string
                    else:
                        v = TypedValue.coerce_to_string(item.proto.data_type, item.proto.data)
                    result.append(v)
                else:
                    result.append((config, item.format_value()))


class AXMLResTablePackage:
    """AXMLResTablePackage"""

    def __init__(self, proto: axml_pb2.AXMLResTablePackage = None):
        """Initialize AXMLResTablePackage

        Args:
            proto (axml_pb2.AXML, optional): _description_. Defaults to None.
        """
        if proto:
            self.proto = proto
        else:
            self.proto = axml_pb2.AXMLResTablePackage()

    @property
    def get_proto(self):
        """get proto of AXMLResTablePackage

        Returns:
            _type_: _description_
        """
        return self.proto

    def compute(self, recursive: bool = True):
        """Compute all fields to have all AXML elements

        Args:
            recursive (bool, optional): need to recompute field recursively. Defaults to True.
        """

        if recursive:
            type_sp_string = StringBlocks(proto=self.proto.type_sp_string)
            type_sp_string.compute(update_size=True)
            self.proto.type_sp_string.CopyFrom(type_sp_string.proto)

            key_sp_string = StringBlocks(proto=self.proto.key_sp_string)
            key_sp_string.compute(update_size=True)
            self.proto.key_sp_string.CopyFrom(key_sp_string.proto)

            for restype in self.proto.restypes:
                if restype.hnd.type == axml_pb2.ResType.RES_TABLE_TYPE_SPEC_TYPE:
                    t = ARSCResTypeSpec(proto=restype.typespec)
                    t.compute(restype.hnd)
                    restype.typespec.CopyFrom(t.proto)
                elif restype.hnd.type == axml_pb2.ResType.RES_TABLE_TYPE_TYPE:
                    t = ARSCResTypeType(proto=restype.typetype)
                    t.compute(restype.hnd)
                    restype.hnd.header_size = 20 + len(
                        ARSCResTableConfig(proto=t.proto.config).pack()
                    )
                    restype.typetype.CopyFrom(t.proto)

        for restype in self.proto.restypes:
            if restype.hnd.type == axml_pb2.ResType.RES_TABLE_TYPE_SPEC_TYPE:
                restype.hnd.size = (
                    len(ARSCResTypeSpec(proto=restype.typespec).pack()) + AXML_HEADER_SIZE
                )
            elif restype.hnd.type == axml_pb2.ResType.RES_TABLE_TYPE_TYPE:
                restype.hnd.size = (
                    len(ARSCResTypeType(proto=restype.typetype).pack()) + AXML_HEADER_SIZE
                )  # - restype.hnd.header_size + AXML_HEADER_SIZE

        # TODO analyse if it is needed
        type_sp_string = StringBlocks(proto=self.proto.type_sp_string).pack()
        self.proto.typeStrings = self.proto.hnd.header_size
        self.proto.keyStrings = self.proto.typeStrings + len(type_sp_string)

        self.proto.hnd.size = len(self.pack())

    def pack(self) -> bytes:
        """pack the AXML element

        Returns:
            bytes: return the AXML element packed
        """
        header = AXMLHeader(proto=self.proto.hnd).pack()
        name = self.proto.name.encode("utf-16")[2:]
        name.ljust(256, b"\x00")
        type_sp_string = StringBlocks(proto=self.proto.type_sp_string).pack()
        key_sp_string = StringBlocks(proto=self.proto.key_sp_string).pack()
        data = b""
        for restype in self.proto.restypes:
            hnd = AXMLHeader(proto=restype.hnd)
            data += hnd.pack()
            if hnd.proto.type == axml_pb2.ResType.RES_TABLE_TYPE_SPEC_TYPE:
                d = ARSCResTypeSpec(proto=restype.typespec).pack()
                data += d
            elif hnd.proto.type == axml_pb2.ResType.RES_TABLE_TYPE_TYPE:
                d = ARSCResTypeType(proto=restype.typetype).pack()
                data += d
        return (
            header
            + pack("<L", self.proto.id)
            + name
            + pack(
                "<LLLL",
                self.proto.typeStrings,
                self.proto.lastPublicType,
                self.proto.keyStrings,
                self.proto.lastPublicKey,
            )
            + self.proto.padding
            + type_sp_string
            + key_sp_string
            + data
        )

    @staticmethod
    def from_axml(buff: bytes, proto=None):
        """Convert AXMLResTablePackage buffer to AXMLResTablePackage object

        Args:
            buff (bytes): buffer contain AXMLResTablePackage object

        Returns:
            tuple[pyaxml.AXMLResTablePackage, bytes]: return AXMLResTablePackage element and buffer offset at the end of the reading
        """
        content = buff
        res_table_package = AXMLResTablePackage()
        if proto:
            res_table_package.proto = proto
        header, buff = AXMLHeader.from_axml(buff, proto=res_table_package.proto.hnd)
        rest = content[header.proto.size :]
        content = content[: header.proto.size]
        # buff = buff[32-8:]
        res_table_package.proto.id = unpack("<L", buff[:4])[0]
        buff = buff[4:]
        res_table_package.proto.name = buff[:256].decode("utf-16")
        buff = buff[256:]
        (
            res_table_package.proto.typeStrings,
            res_table_package.proto.lastPublicType,
            res_table_package.proto.keyStrings,
            res_table_package.proto.lastPublicKey,
        ) = unpack("<LLLL", buff[: 4 * 4])
        buff = buff[4 * 4 :]

        t_str_off = len(content[res_table_package.proto.typeStrings :])
        k_str_off = len(content[res_table_package.proto.keyStrings :])
        if t_str_off > k_str_off:
            len_pad = len(buff) - t_str_off
        else:
            len_pad = len(buff) - k_str_off
        res_table_package.proto.padding = buff[:len_pad]

        _, at = StringBlocks.from_axml(
            content[res_table_package.proto.typeStrings :],
            proto=res_table_package.proto.type_sp_string,
        )

        _, ak = StringBlocks.from_axml(
            content[res_table_package.proto.keyStrings :],
            proto=res_table_package.proto.key_sp_string,
        )

        if len(ak) < len(at):
            content = ak
        else:
            content = at

        while len(content) > 0:
            spec = axml_pb2.ARSCResType()
            hdr, _ = AXMLHeader.from_axml(content, proto=spec.hnd)
            if hdr.proto.type == axml_pb2.ResType.RES_TABLE_TYPE_SPEC_TYPE:
                _, _ = ARSCResTypeSpec.from_axml(content[8 : hdr.proto.size], proto=spec.typespec)
                res_table_package.proto.restypes.append(spec)
            elif hdr.proto.type == axml_pb2.ResType.RES_TABLE_TYPE_TYPE:
                _, _ = ARSCResTypeType.from_axml(content[8 : hdr.proto.size], proto=spec.typetype)
                res_table_package.proto.restypes.append(spec)
            else:
                print("other types ??")
            content = content[hdr.proto.size :]

        return res_table_package, rest

    def set_spec_entry(self, id_: int, entry: int, index: int):
        """Add a spec entry

        Args:
            id_ (int): id of spec entry
            entry (int): _description_
            index (int): index in typetype stringblock of the element to set
        """
        for r in self.proto.restypes:
            if r.hnd.type == axml_pb2.ResType.RES_TABLE_TYPE_SPEC_TYPE and r.typespec.id == id_:
                if len(r.typespec.entries) < index:
                    r.typespec.entries[index] = entry
                else:
                    while index > len(r.typespec.entries):
                        r.typespec.entries.append(0)
                    r.typespec.entries.append(entry)

    def add_id_public(self, type_: str, name: str, id_path: int):
        """Add a public id in ARSC

        Args:
            type_ (str): type of public element
            name (str): name of public element
            id_path (int): id of element value to add in public

        Returns:
            _type_: return the id of this element
        """
        for r in self.proto.restypes:
            if r.hnd.type == axml_pb2.ResType.RES_TABLE_TYPE_TYPE:
                t = StringBlocks(proto=self.proto.type_sp_string).decode_str(r.typetype.id - 1)
                if t == type_:
                    st_key = StringBlocks(proto=self.proto.key_sp_string)
                    spec = ARSCResTableEntry.create_element(st_key.get(name), id_path)
                    self.proto.key_sp_string.CopyFrom(st_key.proto)
                    self.set_spec_entry(r.typetype.id, 0, len(r.typetype.tables))
                    r.typetype.tables.append(spec.proto)
                    return ARSC.convert_id(r.typetype.id, len(r.typetype.tables) - 1)

        st_type = StringBlocks(proto=self.proto.type_sp_string)
        id_ = st_type.get(type_)
        typetype = ARSCResTypeType.create_element(id_=id_ + 1)
        r = axml_pb2.ARSCResType()
        r.hnd.CopyFrom(AXMLHeader(axml_pb2.ResType.RES_TABLE_TYPE_TYPE).proto)
        # compute header without need to do it recursively
        r.hnd.header_size = 20 + len(
            ARSCResTableConfig(proto=typetype.proto.config).pack()
        )
        typetype.compute(r.hnd)
        r.typetype.CopyFrom(typetype.proto)

        st_key = StringBlocks(proto=self.proto.key_sp_string)
        spec = ARSCResTableEntry.create_element(st_key.get(name), id_path)
        self.proto.key_sp_string.CopyFrom(st_key.proto)
        r.typetype.tables.append(spec.proto)

        r_spec = axml_pb2.ARSCResType()
        r_spec.hnd.CopyFrom(AXMLHeader(axml_pb2.ResType.RES_TABLE_TYPE_SPEC_TYPE).proto)
        r_spec.hnd.header_size = 16
        r_spec.typespec.id = id_ + 1

        self.proto.restypes.append(r_spec)
        self.proto.restypes.append(r)
        self.set_spec_entry(id_ + 1, 0, 0)
        return ARSC.convert_id(r.typetype.id, len(r.typetype.tables) - 1)



import unittest
from struct import pack, unpack
import ctypes
from pyaxml.header import (
    TypedValue,
    AXMLHeader,
    AXML_HEADER_SIZE,
    AXMLHeaderXML,
    AXML_RES_TABLE_HEADER_SIZE,
    AXMLHeaderRESTABLE,
)


class TestTypedValue(unittest.TestCase):
    def test_complex_to_float(self):
        # Testing conversion of a complex value to float
        complex_value = 0x00180080  # Example complex value
        expected_float = 6144
        self.assertAlmostEqual(
            TypedValue.complex_to_float(complex_value), expected_float
        )

    def test_coerce_to_string(self):
        # Testing coercion to string for different types
        self.assertEqual(TypedValue.coerce_to_string(TypedValue.TYPE_NULL, 0), "nil")
        self.assertEqual(
            TypedValue.coerce_to_string(TypedValue.TYPE_REFERENCE, 0x1234), "@1234"
        )
        self.assertEqual(
            TypedValue.coerce_to_string(TypedValue.TYPE_ATTRIBUTE, 0x5678), "?5678"
        )
        self.assertEqual(
            TypedValue.coerce_to_string(TypedValue.TYPE_INT_HEX, 0xABCDEF), "0xabcdef"
        )
        self.assertEqual(
            TypedValue.coerce_to_string(TypedValue.TYPE_INT_BOOLEAN, 1), "true"
        )
        self.assertEqual(
            TypedValue.coerce_to_string(TypedValue.TYPE_INT_BOOLEAN, 0), "false"
        )


class TestAXMLHeader(unittest.TestCase):
    def test_pack(self):
        # Create an AXMLHeader instance
        header = AXMLHeader(type_=0x0001, size=16)
        packed_data = header.pack()
        # Verify the packed data
        expected_data = pack("<HHL", 0x0001, AXML_HEADER_SIZE, 16)
        self.assertEqual(packed_data, expected_data)

    def test_from_axml(self):
        # Create a packed buffer
        packed_data = pack("<HHL", 0x0001, AXML_HEADER_SIZE, 16) + b"extra_data"
        # Parse the buffer
        header, remaining_data = AXMLHeader.from_axml(packed_data)
        # Verify the parsed data
        self.assertEqual(header.proto.type, 0x0001)
        self.assertEqual(header.proto.header_size, AXML_HEADER_SIZE)
        self.assertEqual(header.proto.size, 16)
        self.assertEqual(remaining_data, b"extra_data")


class TestAXMLHeaderXML(unittest.TestCase):
    def test_from_axml_valid(self):
        # Create a valid buffer
        packed_data = pack("<HHL", 0x0003, AXML_HEADER_SIZE, 32) + b"extra_data"
        # Parse the buffer
        header_xml, remaining_data = AXMLHeaderXML.from_axml(packed_data)
        # Verify the parsed data
        self.assertEqual(header_xml.proto.type, 0x0003)
        self.assertEqual(header_xml.proto.header_size, AXML_HEADER_SIZE)
        self.assertEqual(header_xml.proto.size, 32)
        self.assertEqual(remaining_data, b"extra_data")

    def test_from_axml_invalid_type(self):
        # Create an invalid buffer
        packed_data = pack("<HHL", 0x0004, AXML_HEADER_SIZE, 32) + b"extra_data"
        # Verify that a TypeError is raised
        with self.assertRaises(TypeError):
            AXMLHeaderXML.from_axml(packed_data)


class TestAXMLHeaderRESTABLE(unittest.TestCase):
    def test_pack(self):
        # Create an AXMLHeaderRESTABLE instance
        header_restable = AXMLHeaderRESTABLE(size=64, package_count=5)
        packed_data = header_restable.pack()
        # Verify the packed data
        expected_data = pack("<HHL", 0x0002, AXML_RES_TABLE_HEADER_SIZE, 64) + pack(
            "<L", 5
        )
        self.assertEqual(packed_data, expected_data)

    def test_from_axml(self):
        # Create a packed buffer
        packed_data = (
            pack("<HHL", 0x0002, AXML_RES_TABLE_HEADER_SIZE, 64)
            + pack("<L", 5)
            + b"extra_data"
        )
        # Parse the buffer
        header_restable, remaining_data = AXMLHeaderRESTABLE.from_axml(packed_data)
        # Verify the parsed data
        self.assertEqual(header_restable.proto.hnd.type, 0x0002)
        self.assertEqual(
            header_restable.proto.hnd.header_size, AXML_RES_TABLE_HEADER_SIZE
        )
        self.assertEqual(header_restable.proto.hnd.size, 64)
        self.assertEqual(header_restable.proto.package_count, 5)
        self.assertEqual(remaining_data, b"extra_data")


if __name__ == "__main__":
    unittest.main()



from struct import pack, unpack
import sys
from typing import Union
import logging

try:
    from pyaxml.proto import axml_pb2
    from pyaxml.header import AXMLHeader, AXML_STRING_POOL_HEADER_SIZE
except ImportError:
    print("proto is not build")
    sys.exit(1)


class AXMLHeaderSTRINGPOOL:
    """AXMLHeaderSTRINGPOOL class to build an AXMLHeaderSTRINGPOOL element"""

    def __init__(
        self,
        sb: Union[list, None] = None,
        size: int = 0,
        proto: axml_pb2.AXMLHeaderSTRINGPOOL = None,
    ):
        """Initialize an AXMLHeader of STRING_POOL

        Args:
            sb (list, optional): list of Stringblock elements. Defaults to None.
            size (int, optional): size of data contain belong to this AXMLHeader. Defaults to 0.
            proto (axml_pb2.AXMLHeaderSTRINGPOOL, optional):
                define AXMLHeaderSTRINGPOOL by a protobuff. Defaults to None.
        """
        # TODO make version with initilisation without proto
        if proto:
            self.proto = proto
        else:
            self.proto = axml_pb2.AXMLHeaderSTRINGPOOL()

    def compute(self):
        """Compute all fields to have a Stringpool element"""

    def pack(self) -> bytes:
        """pack the AXMLHeaderSTRINGPOOL element

        Returns:
            bytes: return the AXMLHeaderSTRINGPOOL element packed
        """
        # TODO add style block for the moment force to 0
        self.proto.len_styleblocks = 0
        return AXMLHeader(proto=self.proto.hnd).pack() + pack(
            "<LLLLL",
            self.proto.len_stringblocks,
            self.proto.len_styleblocks,
            self.proto.flag,
            self.proto.stringoffset,
            self.proto.styleoffset,
        )

    @staticmethod
    def from_axml(buff: bytes, proto=None):
        """Convert AXMLHeaderSTRINGPOOL buffer to AXMLHeaderSTRINGPOOL object

        Args:
            buff (bytes): buffer contain AXMLHeaderSTRINGPOOL object

        Returns:
            tuple[pyaxml.AXMLHeaderSTRINGPOOL, bytes]:
                return AXMLHeaderSTRINGPOOL element and buffer offset at the end of the reading
        """
        hnd_str_pool = AXMLHeaderSTRINGPOOL()
        if proto:
            hnd_str_pool.proto = proto
        _, buff = AXMLHeader.from_axml(buff, proto=hnd_str_pool.proto.hnd)
        (
            hnd_str_pool.proto.len_stringblocks,
            hnd_str_pool.proto.len_styleblocks,
            hnd_str_pool.proto.flag,
            hnd_str_pool.proto.stringoffset,
            hnd_str_pool.proto.styleoffset,
        ) = unpack("<LLLLL", buff[: 5 * 4])
        return hnd_str_pool, buff[5 * 4 :]


##############################################################################
#
#                              STRINGBLOCKS
#
##############################################################################


class StringBlock:
    """StringBlock class to build an StringBlock element"""

    def __init__(
        self,
        data: str = "",
        size: int = 0,
        utf8: bool = False,
        proto: axml_pb2.StringBlock = None,
    ):
        """initialize a stringblock element

        Args:
            data (str, optional): value of the stringblock. Defaults to "".
            size (int, optional): size of data contain belong to this Stringblock. Defaults to 0.
            utf8 (bool, optional): Stringblock can be encoded in UTF8 or UTF16.
                 set True if you want to encode in UTF8 else UTF16. Defaults to False.
            proto (axml_pb2.StringBlock, optional): define StringBlock by a protobuff.
                 Defaults to None.
        """
        self.not_decoded = False
        if proto:
            self.proto = proto
            self.utf8 = utf8
        else:
            self.proto = axml_pb2.StringBlock()
            self.utf8 = utf8
            if self.utf8:
                self.proto.data = data.encode("utf-8")
                self.proto.padding = b"\x00"
            else:
                self.proto.data = data.encode("utf-16")[2:]
                self.proto.padding = b"\x00\x00"
            self.proto.size = size

    def compute(self):
        """Compute all fields to have a StringBlock element"""
        _ = self.proto.size
        if self.utf8:
            try:
                size = len(self.proto.data)

                # print(self.proto.data, hex(self.proto.size))
                decoded_size = len(self.proto.data.decode())
                # if (decoded_size | size << 8) != self.proto.size:
                #    print(hex(size), hex(decoded_size | size << 8), hex(self.proto.size), self.proto.data)
                if size > 0x7F:
                    self.proto.size = (
                        (decoded_size & 0xFF) << 8 | (decoded_size & 0xFF00) >> 8 | 0x80
                    )
                    self.proto.redundant_size = (size & 0xFF) << 8 | (size & 0xFF00) >> 8 | 0x80
                else:
                    if size != decoded_size:
                        self.proto.size = decoded_size | size << 8
                    else:
                        self.proto.size = size | size << 8
            except UnicodeDecodeError:
                logging.error(self.proto.data)
                logging.error("error during decode couldn't recompute Stringblock")
        else:
            size = int(len(self.proto.data) / 2)
            if size > 0x7FFF:
                self.proto.size = (size >> 16) | 0x8000
                self.proto.redundant_size = size & 0xFFFF
            else:
                self.proto.size = size

    def pack(self) -> bytes:
        """pack the StringBlock element

        Returns:
            bytes: return the StringBlock element packed
        """
        redundant = b""
        if self.proto.HasField("redundant_size"):
            redundant = pack("<H", self.proto.redundant_size)
        if self.utf8:
            return pack("<H", self.proto.size) + redundant + self.proto.data + self.proto.padding
        return pack("<H", self.proto.size) + redundant + self.proto.data + self.proto.padding

    @staticmethod
    def from_axml(buff: bytes, utf8: bool, proto=None):
        """Convert StringBlock buffer to StringBlock object

        Args:
            buff (bytes): buffer contain StringBlock object
            utf8 (bool) : specify the encoding of string

        Returns:
            tuple[pyaxml.StringBlock, bytes]:
                 return StringBlock element and buffer offset at the end of the reading
        """
        str_block = StringBlock()
        if proto:
            str_block.proto = proto
        str_block.utf8 = utf8
        if str_block.utf8:
            str_block.proto.size = unpack("<H", buff[:2])[0]
            size = str_block.proto.size >> 8
            # if (str_block.proto.size >> 8) & 0x80 != 0:
            #    print(hex(str_block.proto.size), (str_block.proto.size >> 8) & 0x80)
            if (str_block.proto.size >> 8) & 0x80 != 0:
                redundant = unpack("<H", buff[2:4])[0]
                size = (redundant & 0xFF00) >> 8 | (redundant & 0xF) << 8
                # print(hex(str_block.proto.size >> 8), hex(size))
                buff = buff[2:]
                str_block.proto.redundant_size = redundant
        else:
            str_block.proto.size = unpack("<H", buff[:2])[0]
            if str_block.proto.size & 0x8000 != 0:
                redundant = unpack("<H", buff[2:4])[0]
                size = ((str_block.proto.size & 0x7FFF) << 16) | redundant
                buff = buff[2:]
                str_block.proto.redundant_size = redundant
            size = str_block.proto.size * 2
            while buff[2 + size] != 0:
                size += 1

        str_block.proto.data = buff[2 : 2 + size]
        str_block.proto.padding = buff[2 + size :]
        # print(str_block.proto.data, buff[2 + size :2 + size +99])

        # if str_block.proto.size == 0x8036:
        #    print(hex(size), hex(str_block.proto.size), (str_block.proto.size >> 8) & 0x80)
        #    print(buff[2 : 2 + size +10])
        #    #sys.exit(1)
        return str_block, b""


class StringBlocks:
    """StringBlocks class to build all StringBlocks elements"""

    def __init__(self, proto: axml_pb2.StringBlocks = None):
        """initialize the bunch of StringBlocks element

        Args:
            proto (axml_pb2.StringBlocks, optional):
                 define Stringblocks by a protobuff. Defaults to None.
        """

        if proto:
            self.proto = proto
        else:
            self.proto = axml_pb2.StringBlocks()

    def compute(self, update_size=False):
        """Compute all fields to have all StringBlocks elements"""

        idx = 0
        del self.proto.stringoffsets[:]
        for s in self.proto.stringblocks:
            self.proto.stringoffsets.append(idx)
            s_obj = StringBlock(
                proto=s,
                utf8=self.proto.hnd.flag & axml_pb2.UTF8_FLAG == axml_pb2.UTF8_FLAG,
            )
            if update_size:
                s_obj.compute()
                s.CopyFrom(s_obj.proto)
            idx += len(s_obj.pack())
        sb = b"".join(
            StringBlock(
                proto=elt,
                utf8=self.proto.hnd.flag & axml_pb2.UTF8_FLAG == axml_pb2.UTF8_FLAG,
            ).pack()
            for elt in self.proto.stringblocks
        )
        self.proto.pad_stringblocks = b"\x00" * (4 - (len(sb) % 4)) if len(sb) % 4 != 0 else b""

        self.proto.hnd.stringoffset = (
            AXML_STRING_POOL_HEADER_SIZE
            + len(b"".join(pack("<I", x) for x in self.proto.stringoffsets))
            + len(b"".join(pack("<I", x) for x in self.proto.styleoffsets))
        )

        if len(self.proto.styleoffsets) > 0:
            self.proto.hnd.styleoffset = (
                self.proto.hnd.stringoffset + len(sb) + len(self.proto.pad_stringblocks)
            )
        # self.proto.hnd.styleoffset = self.proto.hnd.stringoffset \
        # + len(self.align(b"".join(StringBlock(proto=elt, \
        # utf8=self.proto.hnd.flag & axml_pb2.UTF8_FLAG == axml_pb2.UTF8_FLAG).pack() \
        #  for elt in self.proto.stringblocks)))
        AXMLHeader(
            axml_pb2.RES_STRING_POOL_TYPE,
            len(self.pack()),
            base_proto=self.proto.hnd.hnd,
        )
        self.proto.hnd.hnd.header_size = AXML_STRING_POOL_HEADER_SIZE
        self.proto.hnd.len_stringblocks = len(self.proto.stringoffsets)
        self.proto.hnd.len_styleblocks = len(self.proto.styleoffsets)

    def align(self, buf: bytes) -> bytes:
        """Align stringblocks elements

        Args:
            buf (bytes): align the buffer given in input

        Returns:
            bytes: return the element with padding to align
        """
        pad = b"\x00" * (4 - (len(buf) % 4))
        if len(pad) == 4:
            return buf
        return buf + pad

    def get(self, name: str) -> int:
        """Get index of a stringblock or if it doesn't exist append a new one.

        Args:
            name (str): the name of the stringblock

        Returns:
            int: return the index of the stringblock
        """
        try:
            index = self.index(name)
        except ValueError:
            index = len(self.proto.stringblocks)
            tmp = StringBlock(
                data=name,
                utf8=self.proto.hnd.flag & axml_pb2.UTF8_FLAG == axml_pb2.UTF8_FLAG,
            )
            tmp.compute()
            self.proto.stringblocks.append(tmp.proto)
        return index

    def index(self, name: str) -> int:
        """Get index of a stringblock or if it doesn't exist raise an error

        Args:
            name (str): the name of the stringblock

        Raises:
            ValueError: raise ValueError if this element didn't exist

        Returns:
            int: return the index of the stringblock
        """
        name_encoded = StringBlock(
            data=name,
            utf8=self.proto.hnd.flag & axml_pb2.UTF8_FLAG == axml_pb2.UTF8_FLAG,
        )
        for i, elt in enumerate(self.proto.stringblocks):
            if elt.data == name_encoded.proto.data:
                return i
        raise ValueError

    def update(self, index: int, name: str) -> None:
        """Update the value of stringblock index with the value name

        Args:
            index (int): index of stringblock
            name (str): value of stringblock
        """
        name_encoded = StringBlock(
            data=name,
            utf8=self.proto.hnd.flag & axml_pb2.UTF8_FLAG == axml_pb2.UTF8_FLAG,
        )
        name_encoded.compute()
        self.proto.stringblocks[index].CopyFrom(name_encoded.proto)

    def replace(self, old_name: str, new_name: str) -> None:
        """Replace a stringblock by a new one

        Args:
            old_name (str): old stringblock
            new_name (str): new stringblock
        """
        index = self.index(old_name)
        self.update(index, new_name)

    def switch(self, name1: str, name2: str) -> None:
        """Switch a name1 stringblock by name2 stringblock

        Args:
            name1 (str): stringblock 1
            name2 (str): stringblock 2
        """
        index1 = self.index(name1)
        index2 = self.index(name2)
        self.update(index1, name2)
        self.update(index2, name1)

    def decode_str(self, index: int) -> str:
        """Decode a stringblock and show its representation in string

        Args:
            index (int): index of stringblock

        Returns:
            str: string representation
        """
        try:
            data = self.proto.stringblocks[index].data
        except IndexError as exc:
            sys.stderr.write(str(self.proto))
            raise IndexError(f"No StringBlock at index: {index}") from exc
        try:
            return (
                data.decode("utf-8")
                if self.proto.hnd.flag & axml_pb2.UTF8_FLAG == axml_pb2.UTF8_FLAG
                else data.decode("utf-16")
            )
        except UnicodeDecodeError:
            return data

    def pack(self) -> bytes:
        """pack the StringBlocks element

        Returns:
            bytes: return the StringBlocks element packed
        """
        sb_offsets = b"".join(pack("<I", x) for x in self.proto.stringoffsets)
        st_offsets = b"".join(pack("<I", x) for x in self.proto.styleoffsets)
        sb = b"".join(
            StringBlock(
                proto=elt,
                utf8=self.proto.hnd.flag & axml_pb2.UTF8_FLAG == axml_pb2.UTF8_FLAG,
            ).pack()
            for elt in self.proto.stringblocks
        )
        sb += self.proto.pad_stringblocks
        st = b""  # TODO
        return AXMLHeaderSTRINGPOOL(proto=self.proto.hnd).pack() + sb_offsets + st_offsets + sb + st

    @staticmethod
    def from_axml(buff: bytes, proto=None):
        """Convert StringBlocks buffer to StringBlocks object

        Args:
            buff (bytes): buffer contain StringBlocks object

        Returns:
            tuple[pyaxml.StringBlocks, bytes]:
              return StringBlocks element and buffer offset at the end of the reading
        """
        rest = buff
        cur_buff = buff
        str_blocks = StringBlocks()
        if proto:
            str_blocks.proto = proto
        hnd, buff = AXMLHeaderSTRINGPOOL.from_axml(buff, proto=str_blocks.proto.hnd)
        buff = buff[: hnd.proto.hnd.size - hnd.proto.hnd.header_size]
        rest = rest[hnd.proto.hnd.size :]
        for i in range(str_blocks.proto.hnd.len_stringblocks):
            str_blocks.proto.stringoffsets.append(unpack("<I", buff[i * 4 : (i + 1) * 4])[0])
        buff = buff[str_blocks.proto.hnd.len_stringblocks * 4 :]
        for i in range(str_blocks.proto.hnd.len_styleblocks):
            # if len(buff[i * 4 : (i + 1) * 4]) != 4:
            #    print(f"WARNING: buff for styleblocks is too small")
            #    return str_blocks, rest
            str_blocks.proto.styleoffsets.append(unpack("<I", buff[i * 4 : (i + 1) * 4])[0])
        buff = buff[str_blocks.proto.hnd.len_styleblocks * 4 :]
        if len(str_blocks.proto.stringoffsets) > 0:
            if str_blocks.proto.hnd.len_styleblocks > 0:
                temp_buff = cur_buff[
                    str_blocks.proto.hnd.stringoffset : str_blocks.proto.hnd.styleoffset
                ]
            else:
                temp_buff = cur_buff[str_blocks.proto.hnd.stringoffset : hnd.proto.hnd.size]
            st_offsets_sort = list(str_blocks.proto.stringoffsets).copy()
            st_offsets_sort.sort()
            for i in range(str_blocks.proto.hnd.len_stringblocks):
                next_index = st_offsets_sort.index(str_blocks.proto.stringoffsets[i]) + 1
                if len(st_offsets_sort) > next_index:
                    next_value = st_offsets_sort[next_index]
                    stbl, _ = StringBlock.from_axml(
                        temp_buff[str_blocks.proto.stringoffsets[i] : next_value],
                        str_blocks.proto.hnd.flag & axml_pb2.UTF8_FLAG == axml_pb2.UTF8_FLAG,
                    )
                else:
                    last_str, _ = StringBlock.from_axml(
                        temp_buff[str_blocks.proto.stringoffsets[i] :],
                        str_blocks.proto.hnd.flag & axml_pb2.UTF8_FLAG == axml_pb2.UTF8_FLAG,
                    )
                    stbl = last_str
                str_blocks.proto.stringblocks.append(stbl.proto)
            buff = buff[str_blocks.proto.stringoffsets[-1] + len(last_str.pack()) :]
            if str_blocks.proto.hnd.len_styleblocks > 0:
                str_blocks.proto.pad_stringblocks = buff[: str_blocks.proto.styleoffsets[0]]
            else:
                str_blocks.proto.pad_stringblocks = buff[: hnd.proto.hnd.size]
        # TODO style
        return str_blocks, rest



import unittest
from struct import pack
from pyaxml.arsctype import ARSCResTableEntry, ARSCComplex, ARSCResStringPoolRef
from pyaxml.proto import axml_pb2


class TestARSCResTableEntry(unittest.TestCase):
    def setUp(self):
        self.proto = axml_pb2.ARSCResTableEntry()

    def test_initialization_with_proto(self):
        obj = ARSCResTableEntry(proto=self.proto)
        self.assertIs(obj.proto, self.proto)

    def test_initialization_without_proto(self):
        obj = ARSCResTableEntry()
        self.assertIsInstance(obj.proto, axml_pb2.ARSCResTableEntry)

    def test_get_proto_property(self):
        obj = ARSCResTableEntry(proto=self.proto)
        self.assertIs(obj.get_proto, self.proto)

    def test_pack_simple_entry(self):
        obj = ARSCResTableEntry(proto=self.proto)
        obj.proto.size = 8
        obj.proto.flags = 0  # Not complex
        obj.proto.index = 42

        packed_data = obj.pack()
        expected_header = pack("<HHI", 8, 0, 42)
        expected_data = b"\x00" * 8
        self.assertEqual(packed_data, expected_header + expected_data)

    def test_pack_complex_entry(self):
        obj = ARSCResTableEntry(proto=self.proto)
        obj.proto.size = 8
        obj.proto.flags = ARSCResTableEntry.FLAG_COMPLEX  # Complex entry
        obj.proto.index = 42

        packed_data = obj.pack()
        expected_header = pack("<HHI", 8, ARSCResTableEntry.FLAG_COMPLEX, 42)
        expected_data = b"\x00" * 8
        self.assertEqual(packed_data, expected_header + expected_data)

    def test_is_public(self):
        obj = ARSCResTableEntry(proto=self.proto)
        obj.proto.flags = ARSCResTableEntry.FLAG_PUBLIC
        self.assertTrue(obj.is_public())

        obj.proto.flags = 0
        self.assertFalse(obj.is_public())

    def test_is_complex(self):
        obj = ARSCResTableEntry(proto=self.proto)
        obj.proto.flags = ARSCResTableEntry.FLAG_COMPLEX
        self.assertTrue(obj.is_complex())

        obj.proto.flags = 0
        self.assertFalse(obj.is_complex())

    def test_is_weak(self):
        obj = ARSCResTableEntry(proto=self.proto)
        obj.proto.flags = ARSCResTableEntry.FLAG_WEAK
        self.assertTrue(obj.is_weak())

        obj.proto.flags = 0
        self.assertFalse(obj.is_weak())

    def test_create_element(self):
        index_name = 42
        data = 100

        obj = ARSCResTableEntry.create_element(index_name, data)
        self.assertEqual(obj.proto.size, 8)
        self.assertEqual(obj.proto.index, index_name)
        self.assertEqual(obj.proto.key.size, 8)
        self.assertEqual(obj.proto.key.data_type, 3)
        self.assertEqual(obj.proto.key.data, data)

    def test_from_axml_simple_entry(self):
        buffer = pack("<HHI", 8, 0, 42) + ARSCResStringPoolRef().pack()
        obj, remaining = ARSCResTableEntry.from_axml(buffer)

        self.assertEqual(obj.proto.size, 8)
        self.assertEqual(obj.proto.flags, 0)
        self.assertEqual(obj.proto.index, 42)
        self.assertEqual(remaining, b"")

    def test_from_axml_complex_entry(self):
        orig = axml_pb2.ARSCResTableEntry()
        orig.size = 8
        orig.flags = ARSCResTableEntry.FLAG_COMPLEX
        orig.index = 42
        orig.item.CopyFrom(ARSCComplex().proto)
        buffer = ARSCResTableEntry(proto=orig).pack()
        obj, remaining = ARSCResTableEntry.from_axml(buffer)

        self.assertEqual(obj.proto.size, 8)
        self.assertEqual(obj.proto.flags, ARSCResTableEntry.FLAG_COMPLEX)
        self.assertEqual(obj.proto.index, 42)
        self.assertEqual(remaining, b"")


if __name__ == "__main__":
    unittest.main()



from struct import pack, unpack
import sys

try:
    from pyaxml.proto import axml_pb2
    from pyaxml.header import TypedValue, AXML_HEADER_SIZE
    from pyaxml.arsctableconfig import ARSCResTableConfig
except ImportError:
    print("proto is not build")
    sys.exit(1)


class ARSCResTypeType:
    """ARSCResTypeType"""

    def __init__(self, proto: axml_pb2.ARSCResTypeType = None):
        """ARSCResTypeType initialize

        Args:
            proto (axml_pb2.ARSCResTypeType, optional): _description_. Defaults to None.
        """
        if proto:
            self.proto = proto
        else:
            self.proto = axml_pb2.ARSCResTypeType()

    @property
    def get_proto(self) -> axml_pb2.ARSCResTypeType:
        """protobuf of ARSCResTypeType

        Returns:
            axml_pb2.ARSCResTypeType: protobuf of ARSCResTypeType
        """
        return self.proto

    def compute(self, hnd: axml_pb2.AXMLHeader, recursive=True):
        """Compute all fields to have all ARSCResTypeType elements

        Args:
            hnd (axml_pb2.AXMLHeader): header of ARSCResTypeType element
            recursive (bool, optional): need to compute all field recursively. Defaults to True.
        """
        cur = 0
        index_cur = 0
        for i in self.proto.tables:
            if not i.present:
                continue
            if len(self.proto.entries) <= index_cur:
                self.proto.entries.append(cur)
            else:
                while self.proto.entries[index_cur] == 0xFFFFFFFF:
                    index_cur += 1
                self.proto.entries[index_cur] = cur
            index_cur += 1
            cur += len(ARSCResTableEntry(proto=i).pack())
        self.proto.entryCount = len(self.proto.entries)
        self.proto.entryStart = hnd.header_size + 4 * self.proto.entryCount

    def pack(self) -> bytes:
        """pack the AXML element

        Returns:
            bytes: return the AXML element packed
        """
        head = pack(
            "<BBHII",
            self.proto.id,
            self.proto.flags,
            self.proto.reserved,
            self.proto.entryCount,
            self.proto.entryStart,
        )
        config = ARSCResTableConfig(proto=self.proto.config).pack()
        data = b""
        for i in self.proto.entries:
            data += pack("<I", i)
        for i in self.proto.tables:
            if not i.present:
                continue
            data += ARSCResTableEntry(proto=i).pack()
        return head + config + data

    @staticmethod
    def from_axml(buff: bytes, proto=None):
        """Convert ARSCResTypeType buffer to ARSCResTypeType object

        Args:
            buff (bytes): buffer contain ARSCResTypeType object

        Returns:
            tuple[pyaxml.ARSCResTypeType, bytes]: return ARSCResTypeType element and buffer offset at the end of the reading
        """
        res_table_typetype = ARSCResTypeType()
        if proto:
            res_table_typetype.proto = proto
        (
            res_table_typetype.proto.id,
            res_table_typetype.proto.flags,
            res_table_typetype.proto.reserved,
            res_table_typetype.proto.entryCount,
            res_table_typetype.proto.entryStart,
        ) = unpack("<BBHII", buff[:12])
        buff = buff[12:]
        _, buff = ARSCResTableConfig.from_axml(buff, proto=res_table_typetype.proto.config)
        for i in range(res_table_typetype.proto.entryCount):
            if len(buff) < 4:
                raise ValueError(
                    f"ARSCResTypeType: entryCount: {res_table_typetype.proto.entryCount} is to big for buffer of size: {len(buff)}"
                )
            res_table_typetype.proto.entries.append(unpack("<I", buff[:4])[0])
            buff = buff[4:]

        entrie_cur = buff
        for i in res_table_typetype.proto.entries:
            if i == 0xFFFFFFFF:
                mock = ARSCResTableEntry()
                mock.proto.present = False
                res_table_typetype.proto.tables.append(mock.proto)
                continue
            table, buff = ARSCResTableEntry.from_axml(entrie_cur[i:])
            res_table_typetype.proto.tables.append(table.proto)
        return res_table_typetype, buff

    @staticmethod
    def create_element(id_: int) -> "ARSCResTypeType":
        """create element ARSCResTypeType

        Args:
            id_ (int): id

        Returns:
            ARSCResTypeType: the ARSCResTypeType created
        """
        typetype = ARSCResTypeType()
        typetype.proto.id = id_
        typetype.proto.flags = 0
        typetype.proto.reserved = 0
        typetype.proto.entryCount = 0
        typetype.proto.config.CopyFrom(ARSCResTableConfig().proto)
        typetype.proto.entryStart = AXML_HEADER_SIZE + 64 + 12
        return typetype


class ARSCResTableEntry:
    """ARSCResTableEntry"""

    # If set, this is a complex entry, holding a set of name/value
    # mappings.  It is followed by an array of ResTable_map structures.
    FLAG_COMPLEX = 1

    # If set, this resource has been declared public, so libraries
    # are allowed to reference it.
    FLAG_PUBLIC = 2

    # If set, this is a weak resource and may be overriden by strong
    # resources of the same name/type. This is only useful during
    # linking with other resource tables.
    FLAG_WEAK = 4

    def __init__(self, proto: axml_pb2.ARSCResTableEntry = None):
        """_summary_

        Args:
            proto (axml_pb2.ARSCResTableEntry, optional): _description_. Defaults to None.
        """
        if proto:
            self.proto = proto
        else:
            self.proto = axml_pb2.ARSCResTableEntry()
            self.proto.present = True

    @property
    def get_proto(self):
        """_summary_

        Returns:
            _type_: _description_
        """
        return self.proto

    def compute(self):
        """Compute all fields to have all ARSCResTableEntry elements"""

    def pack(self) -> bytes:
        """pack the AXML element

        Returns:
            bytes: return the AXML element packed
        """
        data = pack("<HHI", self.proto.size, self.proto.flags, self.proto.index)
        if self.is_complex():
            data += ARSCComplex(proto=self.proto.item).pack()
        else:
            data += ARSCResStringPoolRef(proto=self.proto.key).pack()

        return data

    def is_public(self) -> bool:
        """check if element is public

        Returns:
            bool: True if element is public
        """
        return (self.proto.flags & self.FLAG_PUBLIC) != 0

    def is_complex(self) -> bool:
        """check if element is complex

        Returns:
            bool: True if element is complex
        """
        return (self.proto.flags & self.FLAG_COMPLEX) != 0

    def is_weak(self) -> bool:
        """check if element is weak

        Returns:
            bool: True if element is weak
        """
        return (self.proto.flags & self.FLAG_WEAK) != 0

    @staticmethod
    def create_element(index_name: int, data: int):
        """Create a new ARSCResTableEntry element

        Args:
            index_name (int): The index element
            data (int): index of data element

        Returns:
            pyaxml.ARSCResTableEntry: _description_
        """
        spec = ARSCResTableEntry()
        spec.proto.size = 8
        spec.proto.index = index_name
        spec.proto.key.size = 8
        spec.proto.key.data_type = 3
        spec.proto.key.data = data
        return spec

    @staticmethod
    def from_axml(buff: bytes, proto=None):
        """Convert ARSCResTableEntry buffer to ARSCResTableEntry object

        Args:
            buff (bytes): buffer contain ARSCResTableEntry object

        Returns:
            tuple[pyaxml.ARSCResTableEntry, bytes]: return ARSCResTableEntry element and buffer offset at the end of the reading
        """
        table = ARSCResTableEntry()
        if proto:
            table.proto = proto
        table.proto.size, table.proto.flags, table.proto.index = unpack("<HHI", buff[:8])
        buff = buff[8:]
        if table.is_complex():
            _, buff = ARSCComplex.from_axml(buff, proto=table.proto.item)
        else:
            # If FLAG_COMPLEX is not set, a Res_value structure will follow
            _, buff = ARSCResStringPoolRef.from_axml(buff, proto=table.proto.key)
        # table.proto.input = buff[8:]
        return table, buff


class ARSCComplex:
    """ARSCComplex"""

    def __init__(self, proto: axml_pb2.ARSCComplex = None):
        """Initialize ARSCComplex element

        Args:
            proto (axml_pb2.ARSCComplex, optional): _description_. Defaults to None.
        """
        if proto:
            self.proto = proto
        else:
            self.proto = axml_pb2.ARSCComplex()

    @property
    def get_proto(self):
        """_summary_

        Returns:
            _type_: _description_
        """
        return self.proto

    def compute(self):
        """Compute all fields to have all AXML elements"""

    def pack(self) -> bytes:
        """pack the AXML element

        Returns:
            bytes: return the AXML element packed
        """
        data = pack("<II", self.proto.id_parent, self.proto.count)
        for item in self.proto.items:
            data += pack("<I", item.id)
            data += ARSCResStringPoolRef(proto=item.data).pack()
        return data

    @staticmethod
    def from_axml(buff: bytes, proto=None) -> tuple["ARSCComplex", bytes]:
        """Convert ARSCComplex buffer to ARSCComplex object

        Args:
            buff (bytes): buffer contain ARSCComplex object

        Returns:
            tuple[pyaxml.ARSCComplex, bytes]: return ARSCComplex element and buffer offset at the end of the reading
        """
        complexe = ARSCComplex()
        if proto:
            complexe.proto = proto

        complexe.proto.id_parent, complexe.proto.count = unpack("<II", buff[:8])
        buff = buff[8:]
        # Parse self.count number of `ResTable_map`
        # these are structs of ResTable_ref and Res_value
        # ResTable_ref is a uint32_t.
        for _ in range(0, complexe.proto.count):
            item = axml_pb2.ItemComplex()
            if len(buff) < 4:
                raise ValueError(
                    f"ARSCComplex is to big {complexe.proto.count} is number of complex for buffer {len(buff)}"
                )
            item.id = unpack("<I", buff[:4])[0]
            buff = buff[4:]
            _, buff = ARSCResStringPoolRef.from_axml(buff, proto=item.data)
            complexe.proto.items.append(item)

        return complexe, buff


class ARSCResTypeSpec:
    """ARSCResTypeSpec"""

    def __init__(self, proto: axml_pb2.ARSCResTypeSpec = None):
        """_summary_

        Args:
            proto (axml_pb2.ARSCResTypeSpec, optional): _description_. Defaults to None.
        """
        if proto:
            self.proto = proto
        else:
            self.proto = axml_pb2.ARSCResTypeSpec()

    @property
    def get_proto(self):
        """_summary_

        Returns:
            _type_: _description_
        """
        return self.proto

    def compute(self, hnd: axml_pb2.AXMLHeader):
        """Compute all fields to have all AXML elements"""
        self.proto.entryCount = len(self.proto.entries)

    def pack(self) -> bytes:
        """pack the AXML element

        Returns:
            bytes: return the AXML element packed
        """
        data = pack(
            "<BBHI",
            self.proto.id,
            self.proto.res0,
            self.proto.res1,
            self.proto.entryCount,
        )
        for entry in self.proto.entries:
            data += pack("<I", entry)
        return data

    @staticmethod
    def from_axml(buff: bytes, proto=None):
        """Convert ARSCResTypeSpec buffer to ARSCResTypeSpec object

        Args:
            buff (bytes): buffer contain ARSCResTypeSpec object

        Returns:
            tuple[pyaxml.ARSCResTypeSpec, bytes]: return ARSCResTypeSpec element and buffer offset at the end of the reading
        """
        res_table_typespec = ARSCResTypeSpec()
        if proto:
            res_table_typespec.proto = proto
        (
            res_table_typespec.proto.id,
            res_table_typespec.proto.res0,
            res_table_typespec.proto.res1,
            res_table_typespec.proto.entryCount,
        ) = unpack("<BBHI", buff[:8])
        buff = buff[8:]
        for _ in range(res_table_typespec.proto.entryCount):
            res_table_typespec.proto.entries.append(unpack("<I", buff[:4])[0])
            buff = buff[4:]

        return res_table_typespec, buff


class ARSCResStringPoolRef:
    """ARSCResStringPoolRef"""

    def __init__(self, proto: axml_pb2.ARSCResStringPoolRef = None):
        """_summary_

        Args:
            proto (axml_pb2.ARSCResStringPoolRef, optional): _description_. Defaults to None.
        """
        if proto:
            self.proto = proto
        else:
            self.proto = axml_pb2.ARSCResStringPoolRef()

    @property
    def get_proto(self):
        """_summary_

        Returns:
            _type_: _description_
        """
        return self.proto

    def compute(self):
        """Compute all fields to have all AXML elements"""

    def pack(self) -> bytes:
        """pack the AXML element

        Returns:
            bytes: return the AXML element packed
        """
        data = pack(
            "<HBBI",
            self.proto.size,
            self.proto.res0,
            self.proto.data_type,
            self.proto.data,
        )
        return data

    @staticmethod
    def from_axml(buff: bytes, proto=None):
        """Convert ARSCResStringPoolRef buffer to ARSCResStringPoolRef object

        Args:
            buff (bytes): buffer contain ARSCResStringPoolRef object

        Returns:
            tuple[pyaxml.ARSCResStringPoolRef, bytes]: return ARSCResStringPoolRef element and buffer offset at the end of the reading
        """
        ref = ARSCResStringPoolRef()
        if proto:
            ref.proto = proto

        ref.proto.size, ref.proto.res0, ref.proto.data_type, ref.proto.data = unpack(
            "<HBBI", buff[:8]
        )
        buff = buff[8:]
        return ref, buff

    def is_reference(self) -> bool:
        """
        Returns True if the Res_value is actually a reference to another resource
        """
        return self.proto.data_type == TypedValue.TYPE_REFERENCE



from struct import pack, unpack
import sys
import ctypes
from typing import Union

try:
    from pyaxml.proto import axml_pb2
except ImportError:
    print("proto is not build")
    sys.exit(1)

AXML_HEADER_SIZE = 8
AXML_STRING_POOL_HEADER_SIZE = 28
AXML_RES_TABLE_HEADER_SIZE = AXML_HEADER_SIZE + 4


class AConfiguration:
    """AConfiguration"""

    # See http://aospxref.com/android-13.0.0_r3/xref/frameworks/native/include/android/configuration.h#56

    ORIENTATION_ANY = 0x0000
    ORIENTATION_PORT = 0x0001
    ORIENTATION_LAND = 0x0002
    ORIENTATION_SQUARE = 0x0003
    TOUCHSCREEN_ANY = 0x0000
    TOUCHSCREEN_NOTOUCH = 0x0001
    TOUCHSCREEN_STYLUS = 0x0002
    TOUCHSCREEN_FINGER = 0x0003
    DENSITY_DEFAULT = 0
    DENSITY_LOW = 120
    DENSITY_MEDIUM = 160
    DENSITY_TV = 213
    DENSITY_HIGH = 240
    DENSITY_XHIGH = 320
    DENSITY_XXHIGH = 480
    DENSITY_XXXHIGH = 640
    DENSITY_ANY = 0xFFFE
    DENSITY_NONE = 0xFFFF
    KEYBOARD_ANY = 0x0000
    KEYBOARD_NOKEYS = 0x0001
    KEYBOARD_QWERTY = 0x0002
    KEYBOARD_12KEY = 0x0003
    NAVIGATION_ANY = 0x0000
    NAVIGATION_NONAV = 0x0001
    NAVIGATION_DPAD = 0x0002
    NAVIGATION_TRACKBALL = 0x0003
    NAVIGATION_WHEEL = 0x0004
    KEYSHIDDEN_ANY = 0x0000
    KEYSHIDDEN_NO = 0x0001
    KEYSHIDDEN_YES = 0x0002
    KEYSHIDDEN_SOFT = 0x0003
    NAVHIDDEN_ANY = 0x0000
    NAVHIDDEN_NO = 0x0001
    NAVHIDDEN_YES = 0x0002
    SCREENSIZE_ANY = 0x00
    SCREENSIZE_SMALL = 0x01
    SCREENSIZE_NORMAL = 0x02
    SCREENSIZE_LARGE = 0x03
    SCREENSIZE_XLARGE = 0x04
    SCREENLONG_ANY = 0x00
    SCREENLONG_NO = 0x1
    SCREENLONG_YES = 0x2
    SCREENROUND_ANY = 0x00
    SCREENROUND_NO = 0x1
    SCREENROUND_YES = 0x2
    WIDE_COLOR_GAMUT_ANY = 0x00
    WIDE_COLOR_GAMUT_NO = 0x1
    WIDE_COLOR_GAMUT_YES = 0x2
    HDR_ANY = 0x00
    HDR_NO = 0x1
    HDR_YES = 0x2
    UI_MODE_TYPE_ANY = 0x00
    UI_MODE_TYPE_NORMAL = 0x01
    UI_MODE_TYPE_DESK = 0x02
    UI_MODE_TYPE_CAR = 0x03
    UI_MODE_TYPE_TELEVISION = 0x04
    UI_MODE_TYPE_APPLIANCE = 0x05
    UI_MODE_TYPE_WATCH = 0x06
    UI_MODE_TYPE_VR_HEADSET = 0x07
    UI_MODE_NIGHT_ANY = 0x00
    UI_MODE_NIGHT_NO = 0x1
    UI_MODE_NIGHT_YES = 0x2
    SCREEN_WIDTH_DP_ANY = 0x0000
    SCREEN_HEIGHT_DP_ANY = 0x0000
    SMALLEST_SCREEN_WIDTH_DP_ANY = 0x0000
    LAYOUTDIR_ANY = 0x00
    LAYOUTDIR_LTR = 0x01
    LAYOUTDIR_RTL = 0x02
    MCC = 0x0001
    MNC = 0x0002
    LOCALE = 0x0004
    TOUCHSCREEN = 0x0008
    KEYBOARD = 0x0010
    KEYBOARD_HIDDEN = 0x0020
    NAVIGATION = 0x0040
    ORIENTATION = 0x0080
    DENSITY = 0x0100
    SCREEN_SIZE = 0x0200
    VERSION = 0x0400
    SCREEN_LAYOUT = 0x0800
    UI_MODE = 0x1000
    SMALLEST_SCREEN_SIZE = 0x2000
    LAYOUTDIR = 0x4000
    SCREEN_ROUND = 0x8000
    COLOR_MODE = 0x10000
    MNC_ZERO = 0xFFFF


class ComplexConsts:
    """ComplexConsts"""
    # Units
    UNIT_PX = 0
    UNIT_DIP = 1
    UNIT_SP = 2
    UNIT_PT = 3
    UNIT_IN = 4
    UNIT_MM = 5
    UNIT_FRACTION = 0
    UNIT_FRACTION_PARENT = 1

    # Radix types
    RADIX_23p0 = 0
    RADIX_16p7 = 1
    RADIX_8p15 = 2
    RADIX_0p23 = 3


class TypedValue:
    """TypedValue"""

    TYPE_NULL = 0x0
    TYPE_REFERENCE = 0x1
    TYPE_ATTRIBUTE = 0x02
    TYPE_STRING = 0x03
    TYPE_FLOAT = 0x04
    TYPE_DIMENSION = 0x05
    TYPE_FRACTION = 0x06
    TYPE_FIRST_INT = 0x10
    TYPE_INT_DEC = 0x10
    TYPE_INT_HEX = 0x11
    TYPE_INT_BOOLEAN = 0x12
    TYPE_FIRST_COLOR_INT = 0x1C
    TYPE_INT_COLOR_ARGB8 = 0x1C
    TYPE_INT_COLOR_RGB8 = 0x1D
    TYPE_INT_COLOR_ARGB4 = 0x1E
    TYPE_INT_COLOR_RGB4 = 0x1F
    TYPE_LAST_COLOR_INT = 0x1F
    TYPE_LAST_INT = 0x1F

    COMPLEX_RADIX_SHIFT = 4
    COMPLEX_RADIX_MASK = 0x3
    COMPLEX_UNIT_SHIFT = 0
    COMPLEX_UNIT_MASK = 0xF

    COMPLEX_MANTISSA_SHIFT = 8
    COMPLEX_MANTISSA_MASK = 0xFFFFFF

    DIMENSION_UNIT_STRS = ["px", "dip", "sp", "pt", "in", "mm"]

    FRACTION_UNIT_STRS = ["%", "%p"]

    @staticmethod
    def get_type(t: int):
        """return type of TypedValue

        Args:
            t (int): UID

        Returns:
            _type_: type link to UID
        """
        return t >> 24

    MANTISSA_MULT = 1.0 / float(1 << COMPLEX_MANTISSA_SHIFT)
    RADIX_MULTS = [
        1.0 * MANTISSA_MULT,
        1.0 / float(1 << 7) * MANTISSA_MULT,
        1.0 / (1 << 15) * MANTISSA_MULT,
        1.0 / float(1 << 23) * MANTISSA_MULT,
    ]

    @staticmethod
    def complex_to_float(complexe: int) -> float:
        """convert complex to float

        Args:
            complexe (int): complex value

        Returns:
            float: float value
        """
        return (
            complexe & (TypedValue.COMPLEX_MANTISSA_MASK << TypedValue.COMPLEX_MANTISSA_SHIFT)
        ) * TypedValue.RADIX_MULTS[
            (complexe >> TypedValue.COMPLEX_RADIX_SHIFT) & TypedValue.COMPLEX_RADIX_MASK
        ]

    @staticmethod
    def float_to_complex(val: str) -> int:
        """Convert float to complex int representation (like Android TypedValue).

        Args:
            value (float): The float value (e.g., 400.000000)

        Returns:
            int: Complex representation
        """
        # --- Match suffix to unit ---
        if val.endswith("%"):
            num = val[:-1]
            unit = ComplexConsts.UNIT_FRACTION
        elif val.endswith("dip"):
            num = val[:-3]
            unit = ComplexConsts.UNIT_DIP
        elif val.endswith("dp"):
            num = val[:-2]
            unit = ComplexConsts.UNIT_DIP
        elif val.endswith("sp"):
            num = val[:-2]
            unit = ComplexConsts.UNIT_SP
        elif val.endswith("px"):
            num = val[:-2]
            unit = ComplexConsts.UNIT_PX
        elif val.endswith("pt"):
            num = val[:-2]
            unit = ComplexConsts.UNIT_PT
        elif val.endswith("in"):
            num = val[:-2]
            unit = ComplexConsts.UNIT_IN
        elif val.endswith("mm"):
            num = val[:-2]
            unit = ComplexConsts.UNIT_MM
        else:
            raise ValueError("Invalid unit suffix in: " + val)

        # --- Convert numeric part ---
        f = float(num)

        if -1.0 < f < 1.0:
            base = int(f * (1 << 23))
            radix = ComplexConsts.RADIX_0p23
        elif -0x100 < f < 0x100:
            base = int(f * (1 << 15))
            radix = ComplexConsts.RADIX_8p15
        elif -0x10000 < f < 0x10000:
            base = int(f * (1 << 7))
            radix = ComplexConsts.RADIX_16p7
        else:
            base = int(f)
            radix = ComplexConsts.RADIX_23p0

        return (base << 8) | (radix << 4) | unit

    @staticmethod
    def coerce_to_string(typet: int, data: int) -> Union[str, None]:
        """coerce to string

        Args:
            typet (int): type of value
            data (int): data link to type

        Returns:
            str: coerce string
        """
        if typet == TypedValue.TYPE_NULL:
            return "nil"
        if typet == TypedValue.TYPE_REFERENCE:
            return "@" + hex(data)[2:]
        if typet == TypedValue.TYPE_ATTRIBUTE:
            return "?" + hex(data)[2:]
        if typet == TypedValue.TYPE_FLOAT:
            return str(ctypes.c_float(data).value)
        if typet == TypedValue.TYPE_DIMENSION:
            return (
                str(TypedValue.complex_to_float(data))
                + TypedValue.DIMENSION_UNIT_STRS[
                    (data >> TypedValue.COMPLEX_UNIT_SHIFT) & TypedValue.COMPLEX_UNIT_MASK
                ]
            )
        if typet == TypedValue.TYPE_FRACTION:
            return (
                str(TypedValue.complex_to_float(data) * 100)
                + TypedValue.FRACTION_UNIT_STRS[
                    (data >> TypedValue.COMPLEX_UNIT_SHIFT) & TypedValue.COMPLEX_UNIT_MASK
                ]
            )
        if typet == TypedValue.TYPE_INT_HEX:
            return hex(data)
        if typet == TypedValue.TYPE_INT_BOOLEAN:
            return "true" if data != 0 else "false"
        if typet >= TypedValue.TYPE_FIRST_COLOR_INT and typet <= TypedValue.TYPE_LAST_COLOR_INT:
            return "#" + hex(data)[2:]
        if typet >= TypedValue.TYPE_FIRST_INT and typet <= TypedValue.TYPE_LAST_INT:
            return str(ctypes.c_int32(data).value)

        return None


class AXMLHeader:
    """AXMLHeader class to build an AXMLHeader"""

    def __init__(
        self,
        type_: int = 0,
        size: int = 0,
        proto: axml_pb2.AXMLHeader = None,
        base_proto: axml_pb2.AXMLHeader = None,
    ):
        """Initialize an AXMLHeader

        Args:
            type_ (int, optional): type_ element from ResType. Defaults to 0.
            size (int, optional): size of data contain belong to this AXMLHeader. Defaults to 0.
            proto (axml_pb2.AXMLHeader, optional):
              define AXMLHeader by a protobuff. Defaults to None.
        """

        if proto is None:
            if base_proto:
                self.proto = base_proto
            else:
                self.proto = axml_pb2.AXMLHeader()
            self.proto.type = type_
            self.proto.size = size
            self.proto.header_size = AXML_HEADER_SIZE
        else:
            self.proto = proto

    def pack(self) -> bytes:
        """pack the AXMLHeader element

        Returns:
            bytes: return the AXMLHeader element packed
        """
        return pack("<HHL", self.proto.type, self.proto.header_size, self.proto.size)

    @staticmethod
    def from_axml(buff: bytes, proto=None):
        """Convert AXMLHeader buffer to AXMLHeader object

        Args:
            buff (bytes): buffer contain AXMLHeader object

        Returns:
            tuple[pyaxml.AXMLHeader, bytes]:
              return AXMLHeader element and buffer offset at the end of the reading
        """
        header = AXMLHeader()
        if proto:
            header.proto = proto
        header.proto.type, header.proto.header_size, header.proto.size = unpack("<HHL", buff[:8])
        return header, buff[8:]


class AXMLHeaderXML(AXMLHeader):
    """AXMLHeaderXML class to build an AXMLHeader with the type RES_XML_TYPE"""

    def __init__(self, size: int = 0, proto: axml_pb2.AXMLHeader = None):
        """Initialize an AXMLHeader with the type RES_XML_TYPE

        Args:
            size (int, optional): size of data contain belong to this AXMLHeader. Defaults to 0.
            proto (axml_pb2.AXMLHeader, optional):
              define AXMLHeader by a protobuff. Defaults to None.
        """
        if proto is None:
            super().__init__(axml_pb2.RES_XML_TYPE, size)
        else:
            self.proto = proto

    @staticmethod
    def from_axml(buff: bytes, proto=None):
        """Convert AXMLHeaderXML buffer to AXMLHeaderXML object

        Args:
            buff (bytes): buffer contain AXMLHeaderXML object

        Returns:
            tuple[pyaxml.AXMLHeaderXML, bytes]:
              return AXMLHeaderXML element and buffer offset at the end of the reading
        """
        header_xml = AXMLHeaderXML()
        if proto:
            header_xml.proto = proto
        header, buff = AXMLHeader.from_axml(buff)
        if header.proto.type != axml_pb2.RES_XML_TYPE:
            raise TypeError("AXMLHeaderXML file wrong format no XML_TYPE")
        header_xml.proto.type = header.proto.type
        header_xml.proto.size = header.proto.size
        header_xml.proto.header_size = header.proto.header_size
        return header_xml, buff


class AXMLHeaderRESTABLE:
    """AXMLHeaderRESTABLE class to build an AXMLHeader with the type RES_RES_TABLE_TYPE"""

    def __init__(
        self,
        size: int = 0,
        package_count: int = 0,
        proto: axml_pb2.AXMLHeaderRESTABLE = None,
    ):
        """Initialize AXMLHeader for RES_TABLE

        Args:
            size (int, optional): size of the RES_HEADER block. Defaults to 0.
            package_count (int, optional): number of package. Defaults to 0.
            proto (axml_pb2.AXMLHeaderRESTABLE, optional):
              protobuff of AXMLHeaderRESTABLE. Defaults to None.
        """

        if proto is None:
            self.proto = axml_pb2.AXMLHeaderRESTABLE()
            self.proto.package_count = package_count
            AXMLHeader(axml_pb2.RES_TABLE_TYPE, size, base_proto=self.proto.hnd)
            self.proto.hnd.header_size = AXML_RES_TABLE_HEADER_SIZE
        else:
            self.proto = proto

    def pack(self) -> bytes:
        """pack the AXMLHeader element

        Returns:
            bytes: return the AXMLHeader element packed
        """
        return AXMLHeader(proto=self.proto.hnd).pack() + pack("<L", self.proto.package_count)

    @staticmethod
    def from_axml(buff: bytes, proto=None):
        """Convert AXMLHeaderRESTABLE buffer to AXMLHeaderRESTABLE object

        Args:
            buff (bytes): buffer contain AXMLHeaderRESTABLE object

        Returns:
            tuple[pyaxml.AXMLHeaderRESTABLE, bytes]:
              return AXMLHeaderRESTABLE element and buffer offset at the end of the reading
        """
        restable_header = AXMLHeaderRESTABLE()
        if proto:
            restable_header.proto = proto
        _, buff = AXMLHeader.from_axml(buff, proto=restable_header.proto.hnd)
        restable_header.proto.package_count = unpack("<L", buff[:4])[0]
        return restable_header, buff[4:]



import unittest
from struct import pack
from pyaxml.arscobject import ARSC
from pyaxml.stringblocks import StringBlocks
from pyaxml.proto import axml_pb2
import os

test_dir = os.path.dirname(os.path.abspath(__file__))

class TestARSC(unittest.TestCase):
    def setUp(self):
        """Set up a basic ARSC instance with real axml_pb2 data."""

        

    def test_full_list_packages(self):
        """Test full list_packages metod
        """
        for filename in ["data/ARSC/example.arsc"]:
            ref = ""
            with open(os.path.join(test_dir, filename) + ".xml", "r") as f:
                ref = f.read() 
            with open(os.path.join(test_dir, filename), "rb") as f:
                # Read AXML
                axml, _ = ARSC.from_axml(f.read())
                xml = axml.list_packages()
                self.assertEqual(xml, ref, "resources not correctly decoded")
        


if __name__ == "__main__":
    unittest.main()



import unittest
from struct import pack, unpack
from pyaxml.arsctype import ARSCResTypeType, ARSCResTableConfig
from pyaxml.proto import axml_pb2
from pyaxml.header import AXMLHeader


class TestARSCResTypeType(unittest.TestCase):
    def setUp(self):
        self.proto = axml_pb2.ARSCResTypeType()

    def test_initialization_with_proto(self):
        obj = ARSCResTypeType(proto=self.proto)
        self.assertIs(obj.proto, self.proto)

    def test_initialization_without_proto(self):
        obj = ARSCResTypeType()
        self.assertIsInstance(obj.proto, axml_pb2.ARSCResTypeType)

    def test_get_proto_property(self):
        obj = ARSCResTypeType(proto=self.proto)
        self.assertIs(obj.get_proto, self.proto)

    def test_compute(self):
        header = AXMLHeader().proto
        obj = ARSCResTypeType(proto=self.proto)
        obj.compute(header)

        self.assertEqual(len(obj.proto.entries), len(obj.proto.tables))
        self.assertEqual(obj.proto.entryCount, len(obj.proto.entries))
        self.assertEqual(
            obj.proto.entryStart, header.header_size + 4 * len(obj.proto.entries)
        )

    def test_pack(self):
        obj = ARSCResTypeType(proto=self.proto)
        obj.proto.entries.append(1)
        obj.proto.entries.append(0xFFFFFFFF)
        obj.proto.id = 1
        obj.proto.flags = 0
        obj.proto.reserved = 0
        obj.proto.entryCount = len(obj.proto.entries)
        obj.proto.entryStart = 100

        packed_data = obj.pack()
        expected_head = pack("<BBHII", 1, 0, 0, 2, 100)
        expected_entries = pack("<I", 1) + pack("<I", 0xFFFFFFFF)
        expected_tables = b"\x00\x01\x02\x03" * len(obj.proto.tables)
        pad = b"\x00" * 16
        self.assertEqual(packed_data, expected_head + pad + expected_entries)

    def test_from_axml(self):
        orig = axml_pb2.ARSCResTypeType()
        orig.id = 1
        orig.flags = 0
        orig.entryCount = 2
        orig.config.CopyFrom(ARSCResTableConfig().proto)
        orig.entries.append(0xFFFFFFFF)
        orig.entries.append(0xFFFFFFFF)
        buffer = ARSCResTypeType(proto=orig).pack()

        obj, _ = ARSCResTypeType.from_axml(buffer)

        self.assertEqual(obj.proto.id, 1)
        self.assertEqual(obj.proto.flags, 0)
        self.assertEqual(obj.proto.entryCount, 2)
        self.assertEqual(len(obj.proto.entries), 2)
        self.assertEqual(obj.proto.entries[0], 0xFFFFFFFF)
        self.assertEqual(obj.proto.entries[1], 0xFFFFFFFF)

    def test_create_element(self):
        obj = ARSCResTypeType.create_element(42)
        self.assertEqual(obj.proto.id, 42)
        self.assertEqual(obj.proto.flags, 0)
        self.assertEqual(obj.proto.entryCount, 0)
        # self.assertIsInstance(obj.proto.config, ARSCResTableConfig)


if __name__ == "__main__":
    unittest.main()
